#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017-2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
Subsytem handling localization-related data communication
"""
import collections
import logging
import struct

from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['Localization', 'LocalizationPacket']

logger = logging.getLogger(__name__)

# A generic location packet contains type and data. When received the data
# may be decoded by the lib.
LocalizationPacket = collections.namedtuple('localizationPacket',
                                            ['type', 'raw_data', 'data'])


class Localization():
    """
    Handle localization-related data communication with the Crazyflie
    """

    # Implemented channels
    POSITION_CH = 0
    GENERIC_CH = 1

    # Location message types for generig channel
    RANGE_STREAM_REPORT = 0
    RANGE_STREAM_REPORT_FP16 = 1
    LPS_SHORT_LPP_PACKET = 2
    EMERGENCY_STOP = 3
    EMERGENCY_STOP_WATCHDOG = 4
    COMM_GNSS_NMEA = 6
    COMM_GNSS_PROPRIETARY = 7
    EXT_POSE = 8
    EXT_POSE_PACKED = 9

    def __init__(self, crazyflie=None):
        """
        Initialize the Extpos object.
        """
        self._cf = crazyflie

        self.receivedLocationPacket = Caller()
        self._cf.add_port_callback(CRTPPort.LOCALIZATION, self._incoming)

    def _incoming(self, packet):
        """
        Callback for data received from the copter.
        """
        if len(packet.data) < 1:
            logger.warning('Localization packet received with incorrect' +
                           'length (length is {})'.format(len(packet.data)))
            return

        pk_type = struct.unpack('<B', packet.data[:1])[0]
        data = packet.data[1:]

        # Decoding the known packet types
        # TODO: more generic decoding scheme?
        decoded_data = None
        if pk_type == self.RANGE_STREAM_REPORT:
            if len(data) % 5 != 0:
                logger.error('Wrong range stream report data lenght')
                return
            decoded_data = {}
            raw_data = data
            for i in range(int(len(data) / 5)):
                anchor_id, distance = struct.unpack('<Bf', raw_data[:5])
                decoded_data[anchor_id] = distance
                raw_data = raw_data[5:]

        pk = LocalizationPacket(pk_type, data, decoded_data)
        self.receivedLocationPacket.call(pk)

    def send_extpos(self, pos):
        """
        Send the current Crazyflie X, Y, Z position. This is going to be
        forwarded to the Crazyflie's position estimator.
        """

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.POSITION_CH
        pk.data = struct.pack('<fff', pos[0], pos[1], pos[2])
        self._cf.send_packet(pk)

    def send_extpose(self, pos, quat):
        """
        Send the current Crazyflie pose (position [x, y, z] and
        attitude quaternion [qx, qy, qz, qw]). This is going to be forwarded
        to the Crazyflie's position estimator.
        """

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.GENERIC_CH
        pk.data = struct.pack('<Bfffffff',
                              self.EXT_POSE,
                              pos[0], pos[1], pos[2],
                              quat[0], quat[1], quat[2], quat[3])
        self._cf.send_packet(pk)

    def send_short_lpp_packet(self, dest_id, data):
        """
        Send ultra-wide-band LPP packet to dest_id
        """

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.GENERIC_CH
        pk.data = struct.pack('<BB', self.LPS_SHORT_LPP_PACKET, dest_id) + data
        self._cf.send_packet(pk)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
An early serial link driver. This could still be used (after some fixing) to
run high-speed CRTP with the Crazyflie. The UART can be run at 2Mbit.
"""
import logging
import queue
import re
import threading

from .crtpstack import CRTPPacket
from .exceptions import WrongUriType
from cflib.crtp.crtpdriver import CRTPDriver

found_serial = True
try:
    import serial
    import serial.tools.list_ports as list_ports
except ImportError:
    found_serial = False

__author__ = 'Bitcraze AB'
__all__ = ['SerialDriver']

logger = logging.getLogger(__name__)

MTU = 32
START_BYTE1 = 0xbc
START_BYTE2 = 0xcf
SYSLINK_RADIO_RAW = 0x00


def compute_cksum(list):
    cksum0, cksum1 = 0, 0
    for i in range(len(list)):
        cksum0 = (cksum0 + list[i]) & 0xff
        cksum1 = (cksum1 + cksum0) & 0xff
    return bytearray([cksum0, cksum1])


class SerialDriver(CRTPDriver):

    def __init__(self):
        CRTPDriver.__init__(self)
        self.ser = None
        self.uri = ''
        self.link_error_callback = None
        self.in_queue = None
        self.out_queue = None
        self._receive_thread = None
        self._send_thread = None
        logger.info('Initialized serial driver.')

    def connect(self, uri, linkQualityCallback, linkErrorCallback):
        # check if the URI is a serial URI
        if not re.search('^serial://', uri):
            raise WrongUriType('Not a serial URI')

        # Check if it is a valid serial URI
        uri_data = re.search('^serial://([-a-zA-Z0-9/.]+)$', uri)
        if not uri_data:
            raise Exception('Invalid serial URI')

        if not found_serial:
            raise Exception('PySerial package is missing')

        device_name = uri_data.group(1)
        devices = self.get_devices()
        if device_name not in devices:
            raise Exception('Could not identify device')
        device = devices[device_name]

        self.uri = uri

        self.link_error_callback = linkErrorCallback

        # Prepare the inter-thread communication queue
        self.in_queue = queue.Queue()
        self.out_queue = queue.Queue(1)

        self.ser = serial.Serial(device, 512000, timeout=1)

        # Launch the comm thread
        self._receive_thread = _SerialReceiveThread(
            self.ser, self.in_queue, linkQualityCallback, linkErrorCallback)
        self._receive_thread.start()
        self._send_thread = _SerialSendThread(
            self.ser, self.out_queue, linkQualityCallback, linkErrorCallback)
        self._send_thread.start()

    def send_packet(self, pk):
        try:
            self.out_queue.put(pk, True, 2)
        except queue.Full:
            if self.link_error_callback:
                self.link_error_callback(
                    'RadioDriver: Could not send packet to copter')

    def receive_packet(self, wait=0):
        try:
            if wait == 0:
                pk = self.in_queue.get(False)
            elif wait < 0:
                pk = self.in_queue.get(True)
            else:
                pk = self.in_queue.get(True, wait)
        except queue.Empty:
            return None
        return pk

    def get_status(self):
        return 'No information available'

    def get_name(self):
        return 'serial'

    def scan_interface(self, address):
        if found_serial:
            devices_names = self.get_devices().keys()
            return [('serial://' + x, '') for x in devices_names]
        else:
            return []

    def close(self):
        self._receive_thread.stop()
        self._send_thread.stop()
        try:
            self._receive_thread.join()
            self._send_thread.join()
        except Exception:
            pass
        self.ser.close()

    def get_devices(self):
        result = {}
        for port in list_ports.comports():
            name = port.name
            # Name is not populated on all systems, fall back on the device
            if not name:
                name = port.device

            result[name] = port.device

        return result


class _SerialReceiveThread(threading.Thread):

    def __init__(self, ser, inQueue, link_quality_callback,
                 link_error_callback):
        """ Create the object """
        threading.Thread.__init__(self)
        self.ser = ser
        self.in_queue = inQueue
        self._stop = False
        self.link_error_callback = link_error_callback

    def stop(self):
        """ Stop the thread """
        self._stop = True

    def run(self):
        """ Run the receiver thread """
        READ_END = bytes([START_BYTE1, START_BYTE2])
        received = bytearray(MTU + 4)
        received_header = memoryview(received)[0:2]
        while not self._stop:
            try:
                r = self.ser.read_until(READ_END)[-2:]
                if len(r) != 2:
                    continue

                if r[0] != START_BYTE1 or r[1] != START_BYTE2:
                    continue

                r = self.ser.readinto(received_header)
                if r != 2:
                    continue

                if not (0 < received_header[1] <= MTU):
                    continue

                expected = received_header[1] + 2

                received_data_chk = memoryview(received)[2:2+expected]
                r = self.ser.readinto(received_data_chk)
                if r != expected:
                    continue

                # NOTE: end is (expected - 2) as the length of the data +2 for
                # the header bytes
                cksum = compute_cksum(memoryview(received)[:expected])
                if cksum[0] != received_data_chk[-2] or \
                        cksum[1] != received_data_chk[-1]:
                    continue

                pk = CRTPPacket(received[2], received[3:expected])
                self.in_queue.put(pk)

            except Exception as e:
                import traceback
                if self.link_error_callback:
                    self.link_error_callback(
                        'Error communicating with the Crazyflie!\n'
                        'Exception:%s\n\n%s' % (e, traceback.format_exc()))


class _SerialSendThread(threading.Thread):

    def __init__(self, ser, outQueue, link_quality_callback,
                 link_error_callback):
        """ Create the object """
        threading.Thread.__init__(self)
        self.ser = ser
        self.out_queue = outQueue
        self._stop = False
        self.link_error_callback = link_error_callback

    def stop(self):
        """ Stop the thread """
        self._stop = True

    def run(self):
        """ Run the sender thread """
        out_data = bytearray(MTU + 6)
        out_data[0:3] = bytearray(
            [START_BYTE1, START_BYTE2, SYSLINK_RADIO_RAW])

        empty_packet = CRTPPacket(header=0xFF)
        empty_packet_data_length = 0
        empty_packet_data = bytearray(7)
        empty_packet_data[0:5] = bytearray(
            [START_BYTE1, START_BYTE2, SYSLINK_RADIO_RAW, 0x01,
             empty_packet.header])
        empty_packet_data[5:7] = compute_cksum(empty_packet_data[2:5])

        while not self._stop:
            try:
                pk = self.out_queue.get(True, timeout=0.0003)
                data = pk.data
                len_data = len(data)
                end_of_payload = 5 + len_data

                out_data[3] = len_data + 1
                out_data[4] = pk.header
                out_data[5:end_of_payload] = data
                out_data[end_of_payload:end_of_payload +
                         2] = compute_cksum(out_data[2:end_of_payload])

                written = self.ser.write(out_data[0:end_of_payload + 2])

            except queue.Empty:
                pk = empty_packet
                len_data = empty_packet_data_length
                written = self.ser.write(empty_packet_data)

            if written != len_data + 7:
                if self.link_error_callback:
                    self.link_error_callback(
                        'SerialDriver: Could only send {:d}B bytes of {:d}B '
                        'packet to Crazyflie'.format(
                            written, len_data + 7)
                    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
Enables flash access to the Crazyflie.

"""
import errno
import logging
import struct
from array import array
from binascii import crc32
from functools import reduce
from threading import Lock

from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['Memory', 'MemoryElement']

# Channels used for the logging port
CHAN_INFO = 0
CHAN_READ = 1
CHAN_WRITE = 2

# Commands used when accessing the Settings port
CMD_INFO_VER = 0
CMD_INFO_NBR = 1
CMD_INFO_DETAILS = 2

EEPROM_TOKEN = b'0xBC'

logger = logging.getLogger(__name__)


class MemoryElement(object):
    """A memory """

    TYPE_I2C = 0
    TYPE_1W = 1
    TYPE_DRIVER_LED = 0x10
    TYPE_LOCO = 0x11
    TYPE_TRAJ = 0x12
    TYPE_LOCO2 = 0x13
    TYPE_LH = 0x14
    TYPE_MEMORY_TESTER = 0x15
    TYPE_DRIVER_LEDTIMING = 0x17

    def __init__(self, id, type, size, mem_handler):
        """Initialize the element with default values"""
        self.id = id
        self.type = type
        self.size = size
        self.mem_handler = mem_handler

    @staticmethod
    def type_to_string(t):
        """Get string representation of memory type"""
        if t == MemoryElement.TYPE_I2C:
            return 'I2C'
        if t == MemoryElement.TYPE_1W:
            return '1-wire'
        if t == MemoryElement.TYPE_DRIVER_LEDTIMING:
            return 'LED memory driver'
        if t == MemoryElement.TYPE_DRIVER_LED:
            return 'LED driver'
        if t == MemoryElement.TYPE_LOCO:
            return 'Loco Positioning'
        if t == MemoryElement.TYPE_TRAJ:
            return 'Trajectory'
        if t == MemoryElement.TYPE_LOCO2:
            return 'Loco Positioning 2'
        if t == MemoryElement.TYPE_LH:
            return 'Lighthouse positioning'
        if t == MemoryElement.TYPE_MEMORY_TESTER:
            return 'Memory tester'
        return 'Unknown'

    def new_data(self, mem, addr, data):
        logger.debug('New data, but not OW mem')

    def __str__(self):
        """Generate debug string for memory"""
        return ('Memory: id={}, type={}, size={}'.format(
            self.id, MemoryElement.type_to_string(self.type), self.size))


class LED:
    """Used to set color/intensity of one LED in the LED-ring"""

    def __init__(self):
        """Initialize to off"""
        self.r = 0
        self.g = 0
        self.b = 0
        self.intensity = 100

    def set(self, r, g, b, intensity=None):
        """Set the R/G/B and optionally intensity in one call"""
        self.r = r
        self.g = g
        self.b = b
        if intensity:
            self.intensity = intensity


class LEDDriverMemory(MemoryElement):
    """Memory interface for using the LED-ring mapped memory for setting RGB
       values for all the LEDs in the ring"""

    def __init__(self, id, type, size, mem_handler):
        """Initialize with 12 LEDs"""
        super(LEDDriverMemory, self).__init__(id=id, type=type, size=size,
                                              mem_handler=mem_handler)
        self._update_finished_cb = None
        self._write_finished_cb = None

        self.leds = []
        for i in range(12):
            self.leds.append(LED())

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            logger.debug(
                "Got new data from the LED driver, but we don't care.")

    def write_data(self, write_finished_cb):
        """Write the saved LED-ring data to the Crazyflie"""
        self._write_finished_cb = write_finished_cb
        data = bytearray()
        for led in self.leds:
            # In order to fit all the LEDs in one radio packet RGB565 is used
            # to compress the colors. The calculations below converts 3 bytes
            # RGB into 2 bytes RGB565. Then shifts the value of each color to
            # LSB, applies the intensity and shifts them back for correct
            # alignment on 2 bytes.
            R5 = ((int)((((int(led.r) & 0xFF) * 249 + 1014) >> 11) & 0x1F) *
                  led.intensity / 100)
            G6 = ((int)((((int(led.g) & 0xFF) * 253 + 505) >> 10) & 0x3F) *
                  led.intensity / 100)
            B5 = ((int)((((int(led.b) & 0xFF) * 249 + 1014) >> 11) & 0x1F) *
                  led.intensity / 100)
            tmp = (int(R5) << 11) | (int(G6) << 5) | (int(B5) << 0)
            data += bytearray((tmp >> 8, tmp & 0xFF))
        self.mem_handler.write(self, 0x00, data, flush_queue=True)

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            self.valid = False
            logger.debug('Updating content of memory {}'.format(self.id))
            # Start reading the header
            self.mem_handler.read(self, 0, 16)

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            logger.debug('Write to LED driver done')
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None


class LEDTimingsDriverMemory(MemoryElement):
    """Memory interface for using the LED-ring mapped memory for setting RGB
       values over time. To upload and run a show sequence of
       the LEDs in the ring"""

    def __init__(self, id, type, size, mem_handler):
        super(LEDTimingsDriverMemory, self).__init__(id=id,
                                                     type=type,
                                                     size=size,
                                                     mem_handler=mem_handler)
        self._update_finished_cb = None
        self._write_finished_cb = None

        self.timings = []

    def add(self, time, rgb, leds=0, fade=False, rotate=0):
        self.timings.append({
            'time': time,
            'rgb': rgb,
            'leds': leds,
            'fade': fade,
            'rotate': rotate
        })

    def write_data(self, write_finished_cb):
        if write_finished_cb is not None:
            self._write_finished_cb = write_finished_cb

        data = []
        for timing in self.timings:
            # In order to fit all the LEDs in one radio packet RGB565 is used
            # to compress the colors. The calculations below converts 3 bytes
            # RGB into 2 bytes RGB565. Then shifts the value of each color to
            # LSB, applies the intensity and shifts them back for correct
            # alignment on 2 bytes.
            R5 = ((int)((((int(timing['rgb']['r']) & 0xFF) * 249 + 1014) >> 11)
                        & 0x1F))
            G6 = ((int)((((int(timing['rgb']['g']) & 0xFF) * 253 + 505) >> 10)
                        & 0x3F))
            B5 = ((int)((((int(timing['rgb']['b']) & 0xFF) * 249 + 1014) >> 11)
                        & 0x1F))
            led = (int(R5) << 11) | (int(G6) << 5) | (int(B5) << 0)
            extra = ((timing['leds']) & 0x0F) | (
                (timing['fade'] << 4) & 0x10) | (
                (timing['rotate'] << 5) & 0xE0)

            if (timing['time'] & 0xFF) != 0 or led != 0 or extra != 0:
                data += [timing['time'] & 0xFF, led >> 8, led & 0xFF, extra]

        data += [0, 0, 0, 0]
        self.mem_handler.write(self, 0x00, bytearray(data), flush_queue=True)

    def write_done(self, mem, addr):
        if mem.id == self.id and self._write_finished_cb:
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None


class I2CElement(MemoryElement):

    def __init__(self, id, type, size, mem_handler):
        super(I2CElement, self).__init__(id=id, type=type, size=size,
                                         mem_handler=mem_handler)
        self._update_finished_cb = None
        self._write_finished_cb = None
        self.elements = {}
        self.valid = False

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            if addr == 0:
                done = False
                # Check for header
                if data[0:4] == EEPROM_TOKEN:
                    logger.debug('Got new data: {}'.format(data))
                    [self.elements['version'],
                     self.elements['radio_channel'],
                     self.elements['radio_speed'],
                     self.elements['pitch_trim'],
                     self.elements['roll_trim']] = struct.unpack('<BBBff',
                                                                 data[4:15])
                    if self.elements['version'] == 0:
                        done = True
                    elif self.elements['version'] == 1:
                        self.datav0 = data
                        self.mem_handler.read(self, 16, 5)
                else:
                    self.valid = False
                    if self._update_finished_cb:
                        self._update_finished_cb(self)
                        self._update_finished_cb = None

            if addr == 16:
                [radio_address_upper, radio_address_lower] = struct.unpack(
                    '<BI', self.datav0[15:16] + data[0:4])
                self.elements['radio_address'] = int(
                    radio_address_upper) << 32 | radio_address_lower

                logger.debug(self.elements)
                data = self.datav0 + data
                done = True

            if done:
                if self._checksum256(data[:len(data) - 1]) == \
                        data[len(data) - 1]:
                    self.valid = True
                if self._update_finished_cb:
                    self._update_finished_cb(self)
                    self._update_finished_cb = None

    def _checksum256(self, st):
        return reduce(lambda x, y: x + y, list(st)) % 256

    def write_data(self, write_finished_cb):
        image = bytearray()
        if self.elements['version'] == 0:
            data = (
                0x00, self.elements['radio_channel'],
                self.elements['radio_speed'],
                self.elements['pitch_trim'], self.elements['roll_trim'])
            image += struct.pack('<BBBff', *data)
        elif self.elements['version'] == 1:
            data = (
                0x01, self.elements['radio_channel'],
                self.elements['radio_speed'],
                self.elements['pitch_trim'], self.elements['roll_trim'],
                self.elements['radio_address'] >> 32,
                self.elements['radio_address'] & 0xFFFFFFFF)
            image += struct.pack('<BBBffBI', *data)
        # Adding some magic:
        image = EEPROM_TOKEN + image
        image += struct.pack('B', self._checksum256(image))

        self._write_finished_cb = write_finished_cb

        self.mem_handler.write(self, 0x00,
                               struct.unpack('B' * len(image), image))

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            self.valid = False
            logger.debug('Updating content of memory {}'.format(self.id))
            # Start reading the header
            self.mem_handler.read(self, 0, 16)

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None


class OWElement(MemoryElement):
    """Memory class with extra functionality for 1-wire memories"""

    element_mapping = {
        1: 'Board name',
        2: 'Board revision',
        3: 'Custom'
    }

    def __init__(self, id, type, size, addr, mem_handler):
        """Initialize the memory with good defaults"""
        super(OWElement, self).__init__(id=id, type=type, size=size,
                                        mem_handler=mem_handler)
        self.addr = addr

        self.valid = False

        self.vid = None
        self.pid = None
        self.name = None
        self.pins = None
        self.elements = {}

        self._update_finished_cb = None
        self._write_finished_cb = None

        self._rev_element_mapping = {}
        for key in list(OWElement.element_mapping.keys()):
            self._rev_element_mapping[OWElement.element_mapping[key]] = key

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            if addr == 0:
                if self._parse_and_check_header(data[0:8]):
                    if self._parse_and_check_elements(data[9:11]):
                        self.valid = True
                        self._update_finished_cb(self)
                        self._update_finished_cb = None
                    else:
                        # We need to fetch the elements, find out the length
                        (elem_ver, elem_len) = struct.unpack('BB', data[8:10])
                        self.mem_handler.read(self, 8, elem_len + 3)
                else:
                    # Call the update if the CRC check of the header fails,
                    # we're done here
                    if self._update_finished_cb:
                        self._update_finished_cb(self)
                        self._update_finished_cb = None
            elif addr == 0x08:
                if self._parse_and_check_elements(data):
                    self.valid = True
                if self._update_finished_cb:
                    self._update_finished_cb(self)
                    self._update_finished_cb = None

    def _parse_and_check_elements(self, data):
        """
        Parse and check the CRC and length of the elements part of the memory
        """
        crc = data[-1]
        test_crc = crc32(data[:-1]) & 0x0ff
        elem_data = data[2:-1]
        if test_crc == crc:
            while len(elem_data) > 0:
                (eid, elen) = struct.unpack('BB', elem_data[:2])
                self.elements[self.element_mapping[eid]] = \
                    elem_data[2:2 + elen].decode('ISO-8859-1')
                elem_data = elem_data[2 + elen:]
            return True
        return False

    def write_done(self, mem, addr):
        if self._write_finished_cb:
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def write_data(self, write_finished_cb):
        # First generate the header part
        header_data = struct.pack('<BIBB', 0xEB, self.pins, self.vid, self.pid)
        header_crc = crc32(header_data) & 0x0ff
        header_data += struct.pack('B', header_crc)

        # Now generate the elements part
        elem = bytearray()
        logger.debug(list(self.elements.keys()))
        for element in reversed(list(self.elements.keys())):
            elem_string = self.elements[element]
            key_encoding = self._rev_element_mapping[element]
            elem += struct.pack('BB', key_encoding, len(elem_string))
            elem += bytearray(elem_string.encode('ISO-8859-1'))

        elem_data = struct.pack('BB', 0x00, len(elem))
        elem_data += elem
        elem_crc = crc32(elem_data) & 0x0ff
        elem_data += struct.pack('B', elem_crc)

        data = header_data + elem_data

        self.mem_handler.write(self, 0x00,
                               struct.unpack('B' * len(data), data))

        self._write_finished_cb = write_finished_cb

    def erase(self, write_finished_cb):
        erase_data = array('B', [0xFF] * 112)
        self.mem_handler.write(self, 0x00,
                               struct.unpack('B' * len(erase_data),
                                             erase_data))

        self._write_finished_cb = write_finished_cb

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            self.valid = False
            logger.debug('Updating content of memory {}'.format(self.id))
            # Start reading the header
            self.mem_handler.read(self, 0, 11)

    def _parse_and_check_header(self, data):
        """Parse and check the CRC of the header part of the memory"""
        (start, self.pins, self.vid, self.pid, crc) = struct.unpack('<BIBBB',
                                                                    data)
        test_crc = crc32(data[:-1]) & 0x0ff
        if start == 0xEB and crc == test_crc:
            return True
        return False

    def __str__(self):
        """Generate debug string for memory"""
        return ('OW {} ({:02X}:{:02X}): {}'.format(
            self.addr, self.vid, self.pid, self.elements))

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None


class AnchorData:
    """Holds data for one anchor"""

    def __init__(self, position=(0.0, 0.0, 0.0), is_valid=False):
        self.position = position
        self.is_valid = is_valid

    def set_from_mem_data(self, data):
        x, y, z, self.is_valid = struct.unpack('<fff?', data)
        self.position = (x, y, z)


class LocoMemory(MemoryElement):
    """Memory interface for accessing data from the Loco Positioning system"""

    SIZE_OF_FLOAT = 4
    MEM_LOCO_INFO = 0x0000
    MEM_LOCO_INFO_LEN = 1
    MEM_LOCO_ANCHOR_BASE = 0x1000
    MEM_LOCO_ANCHOR_PAGE_SIZE = 0x0100
    MEM_LOCO_PAGE_LEN = (3 * SIZE_OF_FLOAT) + 1

    def __init__(self, id, type, size, mem_handler):
        super(LocoMemory, self).__init__(id=id, type=type, size=size,
                                         mem_handler=mem_handler)
        self._update_finished_cb = None

        self.anchor_data = []
        self.nr_of_anchors = 0
        self.valid = False

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        done = False
        if mem.id == self.id:
            if addr == LocoMemory.MEM_LOCO_INFO:
                self.nr_of_anchors = data[0]
                if self.nr_of_anchors == 0:
                    done = True
                else:
                    self.anchor_data = \
                        [AnchorData() for _ in range(self.nr_of_anchors)]
                    self._request_page(0)
            else:
                page = int((addr - LocoMemory.MEM_LOCO_ANCHOR_BASE) /
                           LocoMemory.MEM_LOCO_ANCHOR_PAGE_SIZE)

                self.anchor_data[page].set_from_mem_data(data)

                next_page = page + 1
                if next_page < self.nr_of_anchors:
                    self._request_page(next_page)
                else:
                    done = True

        if done:
            self.valid = True
            if self._update_finished_cb:
                self._update_finished_cb(self)
                self._update_finished_cb = None

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            self.anchor_data = []
            self.nr_of_anchors = 0
            self.valid = False
            logger.debug('Updating content of memory {}'.format(self.id))

            # Start reading the header
            self.mem_handler.read(self, LocoMemory.MEM_LOCO_INFO,
                                  LocoMemory.MEM_LOCO_INFO_LEN)

    def disconnect(self):
        self._update_finished_cb = None

    def _request_page(self, page):
        addr = LocoMemory.MEM_LOCO_ANCHOR_BASE + \
            LocoMemory.MEM_LOCO_ANCHOR_PAGE_SIZE * page
        self.mem_handler.read(self, addr, LocoMemory.MEM_LOCO_PAGE_LEN)


class AnchorData2:
    """Holds data for one anchor"""

    def __init__(self, position=(0.0, 0.0, 0.0), is_valid=False):
        self.position = position
        self.is_valid = is_valid

    def set_from_mem_data(self, data):
        x, y, z, self.is_valid = struct.unpack('<fff?', data)
        self.position = (x, y, z)


class LocoMemory2(MemoryElement):
    """Memory interface for accessing data from the Loco Positioning system
       version 2"""

    SIZE_OF_FLOAT = 4

    # MAX_NR_OF_ANCHORS should be set to the number of anchors
    # supported by the firmware. Preferably short enough to fit into one packet
    MAX_NR_OF_ANCHORS = 16
    ID_LIST_LEN = 1 + MAX_NR_OF_ANCHORS

    ADR_ID_LIST = 0x0000
    ADR_ACTIVE_ID_LIST = 0x1000
    ADR_ANCHOR_BASE = 0x2000

    ANCHOR_PAGE_SIZE = 0x0100
    PAGE_LEN = (3 * SIZE_OF_FLOAT) + 1

    def __init__(self, id, type, size, mem_handler):
        super(LocoMemory2, self).__init__(id=id, type=type, size=size,
                                          mem_handler=mem_handler)
        self._update_ids_finished_cb = None
        self._update_active_ids_finished_cb = None
        self._update_data_finished_cb = None
        self._currently_fetching_index = -1

        self.anchor_ids = []
        self.active_anchor_ids = []
        self.anchor_data = {}
        self.nr_of_anchors = 0
        self.ids_valid = False
        self.active_ids_valid = False
        self.data_valid = False

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            if addr == LocoMemory2.ADR_ID_LIST:
                self._handle_id_list_data(data)
            elif addr == LocoMemory2.ADR_ACTIVE_ID_LIST:
                self._handle_active_id_list_data(data)
            else:
                id = int((addr - LocoMemory2.ADR_ANCHOR_BASE) /
                         LocoMemory2.ANCHOR_PAGE_SIZE)
                self._handle_anchor_data(id, data)

    def update_id_list(self, update_ids_finished_cb):
        """Request an update of the id list"""
        if not self._update_ids_finished_cb:
            self._update_ids_finished_cb = update_ids_finished_cb
            self.anchor_ids = []
            self.active_anchor_ids = []
            self.anchor_data = {}

            self.nr_of_anchors = 0
            self.ids_valid = False
            self.data_valid = False

            logger.debug('Updating ids of memory {}'.format(self.id))

            # Start reading the header
            self.mem_handler.read(self, LocoMemory2.ADR_ID_LIST,
                                  LocoMemory2.ID_LIST_LEN)

    def update_active_id_list(self, update_active_ids_finished_cb):
        """Request an update of the active id list"""
        if not self._update_active_ids_finished_cb:
            self._update_active_ids_finished_cb = update_active_ids_finished_cb
            self.active_anchor_ids = []

            self.active_ids_valid = False

            logger.debug('Updating active ids of memory {}'.format(self.id))

            # Start reading the header
            self.mem_handler.read(self, LocoMemory2.ADR_ACTIVE_ID_LIST,
                                  LocoMemory2.ID_LIST_LEN)

    def update_data(self, update_data_finished_cb):
        """Request an update of the anchor data"""
        if not self._update_data_finished_cb and self.nr_of_anchors > 0:
            self._update_data_finished_cb = update_data_finished_cb
            self.anchor_data = {}

            self.data_valid = False
            self._nr_of_anchors_to_fetch = self.nr_of_anchors

            logger.debug('Updating anchor data of memory {}'.format(self.id))

            # Start reading the first anchor
            self._currently_fetching_index = 0
            self._request_page(self.anchor_ids[self._currently_fetching_index])

    def disconnect(self):
        self._update_ids_finished_cb = None
        self._update_data_finished_cb = None

    def _handle_id_list_data(self, data):
        self.nr_of_anchors = data[0]
        for i in range(self.nr_of_anchors):
            self.anchor_ids.append(data[1 + i])
        self.ids_valid = True

        if self._update_ids_finished_cb:
            self._update_ids_finished_cb(self)
            self._update_ids_finished_cb = None

    def _handle_active_id_list_data(self, data):
        count = data[0]
        for i in range(count):
            self.active_anchor_ids.append(data[1 + i])
        self.active_ids_valid = True

        if self._update_active_ids_finished_cb:
            self._update_active_ids_finished_cb(self)
            self._update_active_ids_finished_cb = None

    def _handle_anchor_data(self, id, data):
        anchor = AnchorData2()
        anchor.set_from_mem_data(data)
        self.anchor_data[id] = anchor

        self._currently_fetching_index += 1
        if self._currently_fetching_index < self.nr_of_anchors:
            self._request_page(self.anchor_ids[self._currently_fetching_index])
        else:
            self.data_valid = True
            if self._update_data_finished_cb:
                self._update_data_finished_cb(self)
                self._update_data_finished_cb = None

    def _request_page(self, page):
        addr = LocoMemory2.ADR_ANCHOR_BASE + \
            LocoMemory2.ANCHOR_PAGE_SIZE * page
        self.mem_handler.read(self, addr, LocoMemory2.PAGE_LEN)


class Poly4D:
    class Poly:
        def __init__(self, values=[0.0] * 8):
            self.values = values

    def __init__(self, duration, x=None, y=None, z=None, yaw=None):
        self.duration = duration
        self.x = x if x else self.Poly()
        self.y = y if y else self.Poly()
        self.z = z if z else self.Poly()
        self.yaw = yaw if yaw else self.Poly()


class TrajectoryMemory(MemoryElement):
    """
    Memory interface for trajectories used by the high level commander
    """

    def __init__(self, id, type, size, mem_handler):
        """Initialize trajectory memory"""
        super(TrajectoryMemory, self).__init__(id=id, type=type, size=size,
                                               mem_handler=mem_handler)
        self._write_finished_cb = None

        # A list of Poly4D objects to write to the Crazyflie
        self.poly4Ds = []

    def write_data(self, write_finished_cb):
        """Write trajectory data to the Crazyflie"""
        self._write_finished_cb = write_finished_cb
        data = bytearray()

        for poly4D in self.poly4Ds:
            data += struct.pack('<ffffffff', *poly4D.x.values)
            data += struct.pack('<ffffffff', *poly4D.y.values)
            data += struct.pack('<ffffffff', *poly4D.z.values)
            data += struct.pack('<ffffffff', *poly4D.yaw.values)
            data += struct.pack('<f', poly4D.duration)

        self.mem_handler.write(self, 0x00, data, flush_queue=True)

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            logger.debug('Write trajectory data done')
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._write_finished_cb = None


class LighthouseBsGeometry:
    """Container for geometry data of one Lighthouse base station"""

    SIZE_FLOAT = 4
    SIZE_VECTOR = 3 * SIZE_FLOAT
    SIZE_GEOMETRY = (1 + 3) * SIZE_VECTOR
    SIZE_DATA = 2 * SIZE_GEOMETRY

    def __init__(self):
        self.origin = [0.0, 0.0, 0.0]
        self.rotation_matrix = [
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
        ]

    def set_from_mem_data(self, data):
        self.origin = self._read_vector(
            data[0 * self.SIZE_VECTOR:1 * self.SIZE_VECTOR])
        self.rotation_matrix = [
            self._read_vector(data[1 * self.SIZE_VECTOR:2 * self.SIZE_VECTOR]),
            self._read_vector(data[2 * self.SIZE_VECTOR:3 * self.SIZE_VECTOR]),
            self._read_vector(data[3 * self.SIZE_VECTOR:4 * self.SIZE_VECTOR]),
        ]

    def add_mem_data(self, data):
        self._add_vector(data, self.origin)
        self._add_vector(data, self.rotation_matrix[0])
        self._add_vector(data, self.rotation_matrix[1])
        self._add_vector(data, self.rotation_matrix[2])

    def _add_vector(self, data, vector):
        data += struct.pack('<fff', vector[0], vector[1], vector[2])

    def _read_vector(self, data):
        x, y, z = struct.unpack('<fff', data)
        return [x, y, z]

    def dump(self):
        print('origin:', self.origin)
        print('rotation matrix: ', self.rotation_matrix)


class LighthouseMemory(MemoryElement):
    """
    Memory interface for lighthouse configuration data
    """

    def __init__(self, id, type, size, mem_handler):
        """Initialize Lighthouse memory"""
        super(LighthouseMemory, self).__init__(id=id, type=type, size=size,
                                               mem_handler=mem_handler)

        self._update_finished_cb = None
        self._write_finished_cb = None

        # Geometry data for two base stations
        self.geometry_data = [
            LighthouseBsGeometry(),
            LighthouseBsGeometry(),
        ]

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            if addr == 0:
                self.geometry_data[0].set_from_mem_data(
                    data[0:LighthouseBsGeometry.SIZE_GEOMETRY])
                self.geometry_data[1].set_from_mem_data(
                    data[LighthouseBsGeometry.SIZE_GEOMETRY:])

                if self._update_finished_cb:
                    self._update_finished_cb(self)
                    self._update_finished_cb = None

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            logger.debug('Updating content of memory {}'.format(self.id))
            self.mem_handler.read(self, 0, LighthouseBsGeometry.SIZE_DATA)

    def write_data(self, write_finished_cb):
        """Write geometry data to the Crazyflie"""
        self._write_finished_cb = write_finished_cb
        data = bytearray()

        for bs in self.geometry_data:
            bs.add_mem_data(data)

        self.mem_handler.write(self, 0x00, data, flush_queue=True)

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            logger.debug('Write of geometry data done')
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None

    def dump(self):
        for data in self.geometry_data:
            data.dump()


class MemoryTester(MemoryElement):
    """
    Memory interface for testing the memory sub system, end to end.

    Usage
    1. To verify reading:
      * Call read_data()
      * Wait for the callback to be called
      * Verify that readValidationSucess is True

    2. To verify writing:
      * Set the parameter 'memTst.resetW' in the CF
      * call write_data()
      * Wait for the callback
      * Read the log var 'memTst.errCntW' from the CF and validate that it
        is 0
    """

    def __init__(self, id, type, size, mem_handler):
        """Initialize Memory tester"""
        super(MemoryTester, self).__init__(id=id, type=type, size=size,
                                           mem_handler=mem_handler)

        self._update_finished_cb = None
        self._write_finished_cb = None

        self.readValidationSucess = True

    def new_data(self, mem, start_address, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            for i in range(len(data)):
                actualValue = struct.unpack('<B', data[i:i + 1])[0]
                expectedValue = (start_address + i) & 0xff

                if (actualValue != expectedValue):
                    address = start_address + i
                    self.readValidationSucess = False
                    logger.error(
                        'Error in data - expected: {}, actual: {}, address:{}',
                        expectedValue, actualValue, address)

                if self._update_finished_cb:
                    self._update_finished_cb(self)
                    self._update_finished_cb = None

    def read_data(self, start_address, size, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            logger.debug('Reading memory {}'.format(self.id))
            self.mem_handler.read(self, start_address, size)

    def write_data(self, start_address, size, write_finished_cb):
        """Write data to the Crazyflie"""
        self._write_finished_cb = write_finished_cb
        data = bytearray()

        for i in range(size):
            value = (start_address + i) & 0xff
            data += struct.pack('<B', value)

        self.mem_handler.write(self, start_address, data, flush_queue=True)

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            logger.debug('Write of data finished')
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None


class _ReadRequest:
    """
    Class used to handle memory reads that will split up the read in multiple
    packets if necessary
    """
    MAX_DATA_LENGTH = 20

    def __init__(self, mem, addr, length, cf):
        """Initialize the object with good defaults"""
        self.mem = mem
        self.addr = addr
        self._bytes_left = length
        self.data = bytearray()
        self.cf = cf

        self._current_addr = addr

    def start(self):
        """Start the fetching of the data"""
        self._request_new_chunk()

    def resend(self):
        logger.debug('Sending write again...')
        self._request_new_chunk()

    def _request_new_chunk(self):
        """
        Called to request a new chunk of data to be read from the Crazyflie
        """
        # Figure out the length of the next request
        new_len = self._bytes_left
        if new_len > _ReadRequest.MAX_DATA_LENGTH:
            new_len = _ReadRequest.MAX_DATA_LENGTH

        logger.debug('Requesting new chunk of {}bytes at 0x{:X}'.format(
            new_len, self._current_addr))

        # Request the data for the next address
        pk = CRTPPacket()
        pk.set_header(CRTPPort.MEM, CHAN_READ)
        pk.data = struct.pack('<BIB', self.mem.id, self._current_addr, new_len)
        reply = struct.unpack('<BBBBB', pk.data[:-1])
        self.cf.send_packet(pk, expected_reply=reply, timeout=1)

    def add_data(self, addr, data):
        """Callback when data is received from the Crazyflie"""
        data_len = len(data)
        if not addr == self._current_addr:
            logger.warning(
                'Address did not match when adding data to read request!')
            return

        # Add the data and calculate the next address to fetch
        self.data += data
        self._bytes_left -= data_len
        self._current_addr += data_len

        if self._bytes_left > 0:
            self._request_new_chunk()
            return False
        else:
            return True


class _WriteRequest:
    """
    Class used to handle memory reads that will split up the read in multiple
    packets in necessary
    """
    MAX_DATA_LENGTH = 25

    def __init__(self, mem, addr, data, cf):
        """Initialize the object with good defaults"""
        self.mem = mem
        self.addr = addr
        self._bytes_left = len(data)
        self._data = data
        self.data = bytearray()
        self.cf = cf

        self._current_addr = addr

        self._sent_packet = None
        self._sent_reply = None

        self._addr_add = 0

    def start(self):
        """Start the fetching of the data"""
        self._write_new_chunk()

    def resend(self):
        logger.debug('Sending write again...')
        self.cf.send_packet(
            self._sent_packet, expected_reply=self._sent_reply, timeout=1)

    def _write_new_chunk(self):
        """
        Called to request a new chunk of data to be read from the Crazyflie
        """
        # Figure out the length of the next request
        new_len = len(self._data)
        if new_len > _WriteRequest.MAX_DATA_LENGTH:
            new_len = _WriteRequest.MAX_DATA_LENGTH

        logger.debug('Writing new chunk of {}bytes at 0x{:X}'.format(
            new_len, self._current_addr))

        data = self._data[:new_len]
        self._data = self._data[new_len:]

        pk = CRTPPacket()
        pk.set_header(CRTPPort.MEM, CHAN_WRITE)
        pk.data = struct.pack('<BI', self.mem.id, self._current_addr)
        # Create a tuple used for matching the reply using id and address
        reply = struct.unpack('<BBBBB', pk.data)
        self._sent_reply = reply
        # Add the data
        pk.data += struct.pack('B' * len(data), *data)
        self._sent_packet = pk
        self.cf.send_packet(pk, expected_reply=reply, timeout=1)

        self._addr_add = len(data)

    def write_done(self, addr):
        """Callback when data is received from the Crazyflie"""
        if not addr == self._current_addr:
            logger.warning(
                'Address did not match when adding data to read request!')
            return

        if len(self._data) > 0:
            self._current_addr += self._addr_add
            self._write_new_chunk()
            return False
        else:
            logger.debug('This write request is done')
            return True


class Memory():
    """Access memories on the Crazyflie"""

    # These codes can be decoded using os.stderror, but
    # some of the text messages will look very strange
    # in the UI, so they are redefined here
    _err_codes = {
        errno.ENOMEM: 'No more memory available',
        errno.ENOEXEC: 'Command not found',
        errno.ENOENT: 'No such block id',
        errno.E2BIG: 'Block too large',
        errno.EEXIST: 'Block already exists'
    }

    def __init__(self, crazyflie=None):
        """Instantiate class and connect callbacks"""
        # Called when new memories have been added
        self.mem_added_cb = Caller()
        # Called when new data has been read
        self.mem_read_cb = Caller()

        self.mem_write_cb = Caller()

        self.cf = crazyflie
        self.cf.add_port_callback(CRTPPort.MEM, self._new_packet_cb)
        self.cf.disconnected.add_callback(self._disconnected)
        self._write_requests_lock = Lock()

        self._clear_state()

    def _clear_state(self):
        self.mems = []
        self._refresh_callback = None
        self._fetch_id = 0
        self.nbr_of_mems = 0
        self._ow_mem_fetch_index = 0
        self._elem_data = ()
        self._read_requests = {}
        self._write_requests = {}
        self._ow_mems_left_to_update = []
        self._getting_count = False

    def _mem_update_done(self, mem):
        """
        Callback from each individual memory (only 1-wire) when reading of
        header/elements are done
        """
        if mem.id in self._ow_mems_left_to_update:
            self._ow_mems_left_to_update.remove(mem.id)

        logger.debug(mem)

        if len(self._ow_mems_left_to_update) == 0:
            if self._refresh_callback:
                self._refresh_callback()
                self._refresh_callback = None

    def get_mem(self, id):
        """Fetch the memory with the supplied id"""
        for m in self.mems:
            if m.id == id:
                return m

        return None

    def get_mems(self, type):
        """Fetch all the memories of the supplied type"""
        ret = ()
        for m in self.mems:
            if m.type == type:
                ret += (m,)

        return ret

    def ow_search(self, vid=0xBC, pid=None, name=None):
        """Search for specific memory id/name and return it"""
        for m in self.get_mems(MemoryElement.TYPE_1W):
            if pid and m.pid == pid or name and m.name == name:
                return m

        return None

    def write(self, memory, addr, data, flush_queue=False):
        """Write the specified data to the given memory at the given address"""
        wreq = _WriteRequest(memory, addr, data, self.cf)
        if memory.id not in self._write_requests:
            self._write_requests[memory.id] = []

        # Workaround until we secure the uplink and change messages for
        # mems to non-blocking
        self._write_requests_lock.acquire()
        if flush_queue:
            self._write_requests[memory.id] = self._write_requests[
                memory.id][:1]
        self._write_requests[memory.id].insert(len(self._write_requests), wreq)
        if len(self._write_requests[memory.id]) == 1:
            wreq.start()
        self._write_requests_lock.release()

        return True

    def read(self, memory, addr, length):
        """
        Read the specified amount of bytes from the given memory at the given
        address
        """
        if memory.id in self._read_requests:
            logger.warning('There is already a read operation ongoing for '
                           'memory id {}'.format(memory.id))
            return False

        rreq = _ReadRequest(memory, addr, length, self.cf)
        self._read_requests[memory.id] = rreq

        rreq.start()

        return True

    def refresh(self, refresh_done_callback):
        """Start fetching all the detected memories"""
        self._refresh_callback = refresh_done_callback
        self._fetch_id = 0
        for m in self.mems:
            try:
                self.mem_read_cb.remove_callback(m.new_data)
                m.disconnect()
            except Exception as e:
                logger.info(
                    'Error when removing memory after update: {}'.format(e))
        self.mems = []

        self.nbr_of_mems = 0
        self._getting_count = False

        logger.debug('Requesting number of memories')
        pk = CRTPPacket()
        pk.set_header(CRTPPort.MEM, CHAN_INFO)
        pk.data = (CMD_INFO_NBR,)
        self.cf.send_packet(pk, expected_reply=(CMD_INFO_NBR,))

    def _disconnected(self, uri):
        """The link to the Crazyflie has been broken. Reset state"""
        self._clear_state()

    def _new_packet_cb(self, packet):
        """Callback for newly arrived packets for the memory port"""
        chan = packet.channel
        cmd = packet.data[0]
        payload = packet.data[1:]

        if chan == CHAN_INFO:
            if cmd == CMD_INFO_NBR:
                self.nbr_of_mems = payload[0]
                logger.info('{} memories found'.format(self.nbr_of_mems))

                # Start requesting information about the memories,
                # if there are any...
                if self.nbr_of_mems > 0:
                    if not self._getting_count:
                        self._getting_count = True
                        logger.debug('Requesting first id')
                        pk = CRTPPacket()
                        pk.set_header(CRTPPort.MEM, CHAN_INFO)
                        pk.data = (CMD_INFO_DETAILS, 0)
                        self.cf.send_packet(pk, expected_reply=(
                            CMD_INFO_DETAILS, 0))
                else:
                    self._refresh_callback()

            if cmd == CMD_INFO_DETAILS:

                # Did we get a good reply, otherwise try again:
                if len(payload) < 5:
                    # Workaround for 1-wire bug when memory is detected
                    # but updating the info crashes the communication with
                    # the 1-wire. Fail by saying we only found 1 memory
                    # (the I2C).
                    logger.error(
                        '-------->Got good count, but no info on mem!')
                    self.nbr_of_mems = 1
                    if self._refresh_callback:
                        self._refresh_callback()
                        self._refresh_callback = None
                    return

                # Create information about a new memory
                # Id - 1 byte
                mem_id = payload[0]
                # Type - 1 byte
                mem_type = payload[1]
                # Size 4 bytes (as addr)
                mem_size = struct.unpack('I', payload[2:6])[0]
                # Addr (only valid for 1-wire?)
                mem_addr_raw = struct.unpack('B' * 8, payload[6:14])
                mem_addr = ''
                for m in mem_addr_raw:
                    mem_addr += '{:02X}'.format(m)

                if (not self.get_mem(mem_id)):
                    if mem_type == MemoryElement.TYPE_1W:
                        mem = OWElement(id=mem_id, type=mem_type,
                                        size=mem_size,
                                        addr=mem_addr, mem_handler=self)
                        self.mem_read_cb.add_callback(mem.new_data)
                        self.mem_write_cb.add_callback(mem.write_done)
                        self._ow_mems_left_to_update.append(mem.id)
                    elif mem_type == MemoryElement.TYPE_I2C:
                        mem = I2CElement(id=mem_id, type=mem_type,
                                         size=mem_size,
                                         mem_handler=self)
                        self.mem_read_cb.add_callback(mem.new_data)
                        self.mem_write_cb.add_callback(mem.write_done)
                    elif mem_type == MemoryElement.TYPE_DRIVER_LED:
                        mem = LEDDriverMemory(id=mem_id, type=mem_type,
                                              size=mem_size, mem_handler=self)
                        logger.debug(mem)
                        self.mem_read_cb.add_callback(mem.new_data)
                        self.mem_write_cb.add_callback(mem.write_done)
                    elif mem_type == MemoryElement.TYPE_LOCO:
                        mem = LocoMemory(id=mem_id, type=mem_type,
                                         size=mem_size, mem_handler=self)
                        logger.debug(mem)
                        self.mem_read_cb.add_callback(mem.new_data)
                    elif mem_type == MemoryElement.TYPE_TRAJ:
                        mem = TrajectoryMemory(id=mem_id, type=mem_type,
                                               size=mem_size, mem_handler=self)
                        logger.debug(mem)
                        self.mem_write_cb.add_callback(mem.write_done)
                    elif mem_type == MemoryElement.TYPE_LOCO2:
                        mem = LocoMemory2(id=mem_id, type=mem_type,
                                          size=mem_size, mem_handler=self)
                        logger.debug(mem)
                        self.mem_read_cb.add_callback(mem.new_data)
                    elif mem_type == MemoryElement.TYPE_LH:
                        mem = LighthouseMemory(id=mem_id, type=mem_type,
                                               size=mem_size, mem_handler=self)
                        logger.debug(mem)
                        self.mem_read_cb.add_callback(mem.new_data)
                        self.mem_write_cb.add_callback(mem.write_done)
                    elif mem_type == MemoryElement.TYPE_MEMORY_TESTER:
                        mem = MemoryTester(id=mem_id, type=mem_type,
                                           size=mem_size, mem_handler=self)
                        logger.debug(mem)
                        self.mem_read_cb.add_callback(mem.new_data)
                        self.mem_write_cb.add_callback(mem.write_done)
                    elif mem_type == MemoryElement.TYPE_DRIVER_LEDTIMING:
                        mem = LEDTimingsDriverMemory(id=mem_id, type=mem_type,
                                                     size=mem_size,
                                                     mem_handler=self)
                        logger.debug(mem)
                        self.mem_read_cb.add_callback(mem.new_data)
                        self.mem_write_cb.add_callback(mem.write_done)
                    else:
                        mem = MemoryElement(id=mem_id, type=mem_type,
                                            size=mem_size, mem_handler=self)
                        logger.debug(mem)
                    self.mems.append(mem)
                    self.mem_added_cb.call(mem)

                    self._fetch_id = mem_id + 1

                if self.nbr_of_mems - 1 >= self._fetch_id:
                    logger.debug(
                        'Requesting information about memory {}'.format(
                            self._fetch_id))
                    pk = CRTPPacket()
                    pk.set_header(CRTPPort.MEM, CHAN_INFO)
                    pk.data = (CMD_INFO_DETAILS, self._fetch_id)
                    self.cf.send_packet(pk, expected_reply=(
                        CMD_INFO_DETAILS, self._fetch_id))
                else:
                    logger.debug(
                        'Done getting all the memories, start reading the OWs')
                    ows = self.get_mems(MemoryElement.TYPE_1W)
                    # If there are any OW mems start reading them, otherwise
                    # we are done
                    for ow_mem in ows:
                        ow_mem.update(self._mem_update_done)
                    if len(ows) == 0:
                        if self._refresh_callback:
                            self._refresh_callback()
                            self._refresh_callback = None

        if chan == CHAN_WRITE:
            id = cmd
            (addr, status) = struct.unpack('<IB', payload[0:5])
            logger.debug(
                'WRITE: Mem={}, addr=0x{:X}, status=0x{}'.format(
                    id, addr, status))
            # Find the read request
            if id in self._write_requests:
                self._write_requests_lock.acquire()
                wreq = self._write_requests[id][0]
                if status == 0:
                    if wreq.write_done(addr):
                        # self._write_requests.pop(id, None)
                        # Remove the first item
                        self._write_requests[id].pop(0)
                        self.mem_write_cb.call(wreq.mem, wreq.addr)

                        # Get a new one to start (if there are any)
                        if len(self._write_requests[id]) > 0:
                            self._write_requests[id][0].start()
                else:
                    logger.debug(
                        'Status {}: write resending...'.format(status))
                    wreq.resend()
                self._write_requests_lock.release()

        if chan == CHAN_READ:
            id = cmd
            (addr, status) = struct.unpack('<IB', payload[0:5])
            data = struct.unpack('B' * len(payload[5:]), payload[5:])
            logger.debug('READ: Mem={}, addr=0x{:X}, status=0x{}, '
                         'data={}'.format(id, addr, status, data))
            # Find the read request
            if id in self._read_requests:
                logger.debug(
                    'READING: We are still interested in request for '
                    'mem {}'.format(id))
                rreq = self._read_requests[id]
                if status == 0:
                    if rreq.add_data(addr, payload[5:]):
                        self._read_requests.pop(id, None)
                        self.mem_read_cb.call(rreq.mem, rreq.addr, rreq.data)
                else:
                    logger.debug('Status {}: resending...'.format(status))
                    rreq.resend()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
Crazyradio CRTP link driver.

This driver is used to communicate with the Crazyflie using the Crazyradio
USB dongle.
"""
import array
import binascii
import collections
import logging
import queue
import re
import struct
import threading

import cflib.drivers.crazyradio as crazyradio
from .crtpstack import CRTPPacket
from .exceptions import WrongUriType
from cflib.crtp.crtpdriver import CRTPDriver
from cflib.drivers.crazyradio import Crazyradio


__author__ = 'Bitcraze AB'
__all__ = ['RadioDriver']

logger = logging.getLogger(__name__)

_nr_of_retries = 100
_nr_of_arc_retries = 3

DEFAULT_ADDR_A = [0xe7, 0xe7, 0xe7, 0xe7, 0xe7]
DEFAULT_ADDR = 0xE7E7E7E7E7


class _SharedRadio():
    """ Manages access to one shared radio
    """

    def __init__(self, devid):
        self.radio = Crazyradio(devid=devid)
        self.lock = threading.Lock()
        self.usage_counter = 0


class _RadioManager():
    """ Radio manager helper class
     Get a Crazyradio with:
       radio_manager = _RadioManager(devid)
     Then use your Crazyradio:
       with radio_manager as cradio:
           # cradio is the Crazyradio driver object, it is locked
     Finally close it when finished.
      cradio.close()
    """
    # Configuration lock. Protects opening and closing Crazyradios
    _config_lock = threading.Lock()

    _radios = []  # Hardware Crazyradio objects

    def __init__(self, devid, channel=0, datarate=0, address=DEFAULT_ADDR_A):
        self._devid = devid
        self._channel = channel
        self._datarate = datarate
        self._address = address

        with _RadioManager._config_lock:
            if len(_RadioManager._radios) <= self._devid or \
                    _RadioManager._radios[self._devid] is None:
                _RadioManager._radios += ((self._devid + 1) -
                                          len(_RadioManager._radios)) * [None]
                _RadioManager._radios[self._devid] = _SharedRadio(self._devid)

            _RadioManager._radios[self._devid].usage_counter += 1

    def close(self):
        with _RadioManager._config_lock:
            _RadioManager._radios[self._devid].usage_counter -= 1

            if _RadioManager._radios[self._devid].usage_counter == 0:
                try:
                    _RadioManager._radios[self._devid].radio.close()
                except Exception:
                    pass
                _RadioManager._radios[self._devid] = None

    def __enter__(self):
        _RadioManager._radios[self._devid].lock.acquire()

        _RadioManager._radios[self._devid].radio.set_channel(self._channel)
        _RadioManager._radios[self._devid].radio.set_data_rate(self._datarate)
        _RadioManager._radios[self._devid].radio.set_address(self._address)

        return _RadioManager._radios[self._devid].radio

    def __exit__(self, type, value, traceback):
        _RadioManager._radios[self._devid].lock.release()


class RadioDriver(CRTPDriver):
    """ Crazyradio link driver """

    def __init__(self):
        """ Create the link driver """
        CRTPDriver.__init__(self)
        self._radio_manager = None
        self.uri = ''
        self.link_error_callback = None
        self.link_quality_callback = None
        self.in_queue = None
        self.out_queue = None
        self._thread = None
        self.needs_resending = True

    def connect(self, uri, link_quality_callback, link_error_callback):
        """
        Connect the link driver to a specified URI of the format:
        radio://<dongle nbr>/<radio channel>/[250K,1M,2M]

        The callback for linkQuality can be called at any moment from the
        driver to report back the link quality in percentage. The
        callback from linkError will be called when a error occurs with
        an error message.
        """

        devid, channel, datarate, address = self.parse_uri(uri)
        self.uri = uri

        if self._radio_manager is None:
            self._radio_manager = _RadioManager(devid,
                                                channel,
                                                datarate,
                                                address)
        else:
            raise Exception('Link already open!')

        with self._radio_manager as cradio:
            if cradio.version >= 0.4:
                cradio.set_arc(_nr_of_arc_retries)
            else:
                logger.warning('Radio version <0.4 will be obsoleted soon!')

        # Prepare the inter-thread communication queue
        self.in_queue = queue.Queue()
        # Limited size out queue to avoid "ReadBack" effect
        self.out_queue = queue.Queue(1)

        # Launch the comm thread
        self._thread = _RadioDriverThread(self._radio_manager,
                                          self.in_queue,
                                          self.out_queue,
                                          link_quality_callback,
                                          link_error_callback,
                                          self)
        self._thread.start()

        self.link_error_callback = link_error_callback

    @staticmethod
    def parse_uri(uri):
        # check if the URI is a radio URI
        if not re.search('^radio://', uri):
            raise WrongUriType('Not a radio URI')

        # Open the USB dongle
        if not re.search('^radio://([0-9a-fA-F]+)((/([0-9]+))'
                         '((/(250K|1M|2M))?(/([A-F0-9]+))?)?)?$', uri):
            raise WrongUriType('Wrong radio URI format!')

        uri_data = re.search('^radio://([0-9a-fA-F]+)((/([0-9]+))'
                             '((/(250K|1M|2M))?(/([A-F0-9]+))?)?)?$', uri)

        if len(uri_data.group(1)) < 10 and uri_data.group(1).isdigit():
            devid = int(uri_data.group(1))
        else:
            try:
                devid = crazyradio.get_serials().index(
                    uri_data.group(1).upper())
            except ValueError:
                raise Exception('Cannot find radio with serial {}'.format(
                    uri_data.group(1)))

        channel = 2
        if uri_data.group(4):
            channel = int(uri_data.group(4))

        datarate = Crazyradio.DR_2MPS
        if uri_data.group(7) == '250K':
            datarate = Crazyradio.DR_250KPS
        if uri_data.group(7) == '1M':
            datarate = Crazyradio.DR_1MPS
        if uri_data.group(7) == '2M':
            datarate = Crazyradio.DR_2MPS

        address = DEFAULT_ADDR_A
        if uri_data.group(9):
            addr = str(uri_data.group(9))
            new_addr = struct.unpack('<BBBBB', binascii.unhexlify(addr))
            address = new_addr

        return devid, channel, datarate, address

    def receive_packet(self, time=0):
        """
        Receive a packet though the link. This call is blocking but will
        timeout and return None if a timeout is supplied.
        """
        if time == 0:
            try:
                return self.in_queue.get(False)
            except queue.Empty:
                return None
        elif time < 0:
            try:
                return self.in_queue.get(True)
            except queue.Empty:
                return None
        else:
            try:
                return self.in_queue.get(True, time)
            except queue.Empty:
                return None

    def send_packet(self, pk):
        """ Send the packet pk though the link """
        try:
            self.out_queue.put(pk, True, 2)
        except queue.Full:
            if self.link_error_callback:
                self.link_error_callback('RadioDriver: Could not send packet'
                                         ' to copter')

    def pause(self):
        self._thread.stop()
        self._thread = None

    def restart(self):
        if self._thread:
            return

        self._thread = _RadioDriverThread(self._radio_manager, self.in_queue,
                                          self.out_queue,
                                          self.link_quality_callback,
                                          self.link_error_callback,
                                          self)
        self._thread.start()

    def close(self):
        """ Close the link. """
        # Stop the comm thread
        self._thread.stop()

        # Close the USB dongle
        if self._radio_manager:
            self._radio_manager.close()
        self._radio_manager = None

        while not self.out_queue.empty():
            self.out_queue.get()

        # Clear callbacks
        self.link_error_callback = None
        self.link_quality_callback = None

    def _scan_radio_channels(self, cradio, start=0, stop=125):
        """ Scan for Crazyflies between the supplied channels. """
        return list(cradio.scan_channels(start, stop, (0xff,)))

    def scan_selected(self, links):
        to_scan = ()
        for link in links:
            one_to_scan = {}
            uri_data = re.search('^radio://([0-9]+)((/([0-9]+))'
                                 '(/(250K|1M|2M))?)?$',
                                 link)

            one_to_scan['channel'] = int(uri_data.group(4))

            datarate = Crazyradio.DR_2MPS
            if uri_data.group(6) == '250K':
                datarate = Crazyradio.DR_250KPS
            if uri_data.group(6) == '1M':
                datarate = Crazyradio.DR_1MPS
            if uri_data.group(6) == '2M':
                datarate = Crazyradio.DR_2MPS

            one_to_scan['datarate'] = datarate

            to_scan += (one_to_scan,)

        with self._radio_manager as cradio:
            found = cradio.scan_selected(to_scan, (0xFF, 0xFF, 0xFF))

        ret = ()
        for f in found:
            dr_string = ''
            if f['datarate'] == Crazyradio.DR_2MPS:
                dr_string = '2M'
            if f['datarate'] == Crazyradio.DR_250KPS:
                dr_string = '250K'
            if f['datarate'] == Crazyradio.DR_1MPS:
                dr_string = '1M'

            ret += ('radio://0/{}/{}'.format(f['channel'], dr_string),)

        return ret

    def scan_interface(self, address):
        """ Scan interface for Crazyflies """

        if self._radio_manager is None:
            try:
                self._radio_manager = _RadioManager(0)
            except Exception:
                return []

        with self._radio_manager as cradio:
            # FIXME: implements serial number in the Crazyradio driver!
            serial = 'N/A'

            logger.info('v%s dongle with serial %s found', cradio.version,
                        serial)
            found = []

            if address is not None:
                addr = '{:X}'.format(address)
                new_addr = struct.unpack('<BBBBB', binascii.unhexlify(addr))
                cradio.set_address(new_addr)

            cradio.set_arc(1)

            cradio.set_data_rate(cradio.DR_250KPS)

            if address is None or address == DEFAULT_ADDR:
                found += [['radio://0/{}/250K'.format(c), '']
                          for c in self._scan_radio_channels(cradio)]
                cradio.set_data_rate(cradio.DR_1MPS)
                found += [['radio://0/{}/1M'.format(c), '']
                          for c in self._scan_radio_channels(cradio)]
                cradio.set_data_rate(cradio.DR_2MPS)
                found += [['radio://0/{}/2M'.format(c), '']
                          for c in self._scan_radio_channels(cradio)]
            else:
                found += [['radio://0/{}/250K/{:X}'.format(c, address), '']
                          for c in self._scan_radio_channels(cradio)]
                cradio.set_data_rate(cradio.DR_1MPS)
                found += [['radio://0/{}/1M/{:X}'.format(c, address), '']
                          for c in self._scan_radio_channels(cradio)]
                cradio.set_data_rate(cradio.DR_2MPS)
                found += [['radio://0/{}/2M/{:X}'.format(c, address), '']
                          for c in self._scan_radio_channels(cradio)]

        self._radio_manager.close()
        self._radio_manager = None

        return found

    def get_status(self):
        try:
            radio_manager = _RadioManager(0)

            with radio_manager as cradio:
                ver = cradio.version

            radio_manager.close()

            return 'Crazyradio version {}'.format(ver)
        except Exception:
            return 'Crazyradio not found'

    def get_name(self):
        return 'radio'


# Transmit/receive radio thread
class _RadioDriverThread(threading.Thread):
    """
    Radio link receiver thread used to read data from the
    Crazyradio USB driver. """

    def __init__(self, radio_manager, inQueue, outQueue,
                 link_quality_callback, link_error_callback, link):
        """ Create the object """
        threading.Thread.__init__(self)
        self._radio_manager = radio_manager
        self._in_queue = inQueue
        self._out_queue = outQueue
        self._sp = False
        self._link_error_callback = link_error_callback
        self._link_quality_callback = link_quality_callback
        self._retry_before_disconnect = _nr_of_retries
        self._retries = collections.deque()
        self._retry_sum = 0

        self._curr_up = 0
        self._curr_down = 1

        self._has_safelink = False
        self._link = link

    def stop(self):
        """ Stop the thread """
        self._sp = True
        try:
            self.join()
        except Exception:
            pass

    def _send_packet_safe(self, cr, packet):
        """
        Adds 1bit counter to CRTP header to guarantee that no ack (downlink)
        payload are lost and no uplink packet are duplicated.
        The caller should resend packet if not acked (ie. same as with a
        direct call to crazyradio.send_packet)
        """
        # packet = bytearray(packet)
        packet[0] &= 0xF3
        packet[0] |= self._curr_up << 3 | self._curr_down << 2
        resp = cr.send_packet(packet)
        if resp and resp.ack and len(resp.data) and \
           (resp.data[0] & 0x04) == (self._curr_down << 2):
            self._curr_down = 1 - self._curr_down
        if resp and resp.ack:
            self._curr_up = 1 - self._curr_up

        return resp

    def run(self):
        """ Run the receiver thread """
        dataOut = array.array('B', [0xFF])
        waitTime = 0
        emptyCtr = 0

        # Try up to 10 times to enable the safelink mode
        with self._radio_manager as cradio:
            for _ in range(10):
                resp = cradio.send_packet((0xff, 0x05, 0x01))
                if resp and resp.data and tuple(resp.data) == (
                        0xff, 0x05, 0x01):
                    self._has_safelink = True
                    self._curr_up = 0
                    self._curr_down = 0
                    break
        self._link.needs_resending = not self._has_safelink

        while (True):
            if (self._sp):
                break

            with self._radio_manager as cradio:
                try:
                    if self._has_safelink:
                        ackStatus = self._send_packet_safe(cradio, dataOut)
                    else:
                        ackStatus = cradio.send_packet(dataOut)
                except Exception as e:
                    import traceback

                    self._link_error_callback(
                        'Error communicating with crazy radio ,it has '
                        'probably been unplugged!\nException:%s\n\n%s' % (
                            e, traceback.format_exc()))

            # Analyse the in data packet ...
            if ackStatus is None:
                logger.info('Dongle reported ACK status == None')
                continue

            if (self._link_quality_callback is not None):
                # track the mean of a sliding window of the last N packets
                retry = 10 - ackStatus.retry
                self._retries.append(retry)
                self._retry_sum += retry
                if len(self._retries) > 100:
                    self._retry_sum -= self._retries.popleft()
                link_quality = float(self._retry_sum) / len(self._retries) * 10
                self._link_quality_callback(link_quality)

            # If no copter, retry
            if ackStatus.ack is False:
                self._retry_before_disconnect = \
                    self._retry_before_disconnect - 1
                if (self._retry_before_disconnect == 0 and
                        self._link_error_callback is not None):
                    self._link_error_callback('Too many packets lost')
                continue
            self._retry_before_disconnect = _nr_of_retries

            data = ackStatus.data

            # If there is a copter in range, the packet is analysed and the
            # next packet to send is prepared
            if (len(data) > 0):
                inPacket = CRTPPacket(data[0], list(data[1:]))
                self._in_queue.put(inPacket)
                waitTime = 0
                emptyCtr = 0
            else:
                emptyCtr += 1
                if (emptyCtr > 10):
                    emptyCtr = 10
                    # Relaxation time if the last 10 packet where empty
                    waitTime = 0.01
                else:
                    waitTime = 0

            # get the next packet to send of relaxation (wait 10ms)
            outPacket = None
            try:
                outPacket = self._out_queue.get(True, waitTime)
            except queue.Empty:
                outPacket = None

            dataOut = array.array('B')

            if outPacket:
                dataOut.append(outPacket.header)
                for X in outPacket.data:
                    if type(X) == int:
                        dataOut.append(X)
                    else:
                        dataOut.append(ord(X))
            else:
                dataOut.append(0xFF)


def set_retries_before_disconnect(nr_of_retries):
    global _nr_of_retries
    _nr_of_retries = nr_of_retries


def set_retries(nr_of_arc_retries):
    global _nr_of_arc_retries
    _nr_of_arc_retries = nr_of_arc_retries



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
The synchronous Crazyflie class is a wrapper around the "normal" Crazyflie
class. It handles the asynchronous nature of the Crazyflie API and turns it
into blocking function. It is useful for simple scripts that performs tasks
as a sequence of events.
"""
from threading import Event

from cflib.crazyflie import Crazyflie


class SyncCrazyflie:

    def __init__(self, link_uri, cf=None):
        """ Create a synchronous Crazyflie instance with the specified
        link_uri """

        if cf:
            self.cf = cf
        else:
            self.cf = Crazyflie()

        self._link_uri = link_uri
        self._connect_event = None
        self._disconnect_event = None
        self._is_link_open = False
        self._error_message = None

    def open_link(self):
        if (self.is_link_open()):
            raise Exception('Link already open')

        self._add_callbacks()

        print('Connecting to %s' % self._link_uri)

        self._connect_event = Event()
        self.cf.open_link(self._link_uri)
        self._connect_event.wait()
        self._connect_event = None

        if not self._is_link_open:
            self._remove_callbacks()
            raise Exception(self._error_message)

    def __enter__(self):
        self.open_link()
        return self

    def close_link(self):
        if (self.is_link_open()):
            self._disconnect_event = Event()
            self.cf.close_link()
            self._disconnect_event.wait()
            self._disconnect_event = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_link()

    def is_link_open(self):
        return self._is_link_open

    def _connected(self, link_uri):
        """ This callback is called form the Crazyflie API when a Crazyflie
        has been connected and the TOCs have been downloaded."""
        print('Connected to %s' % link_uri)
        self._is_link_open = True
        if self._connect_event:
            self._connect_event.set()

    def _connection_failed(self, link_uri, msg):
        """Callback when initial connection fails (i.e no Crazyflie
        at the specified address)"""
        print('Connection to %s failed: %s' % (link_uri, msg))
        self._is_link_open = False
        self._error_message = msg
        if self._connect_event:
            self._connect_event.set()

    def _disconnected(self, link_uri):
        self._remove_callbacks()
        self._is_link_open = False
        if self._disconnect_event:
            self._disconnect_event.set()

    def _add_callbacks(self):
        self.cf.connected.add_callback(self._connected)
        self.cf.connection_failed.add_callback(self._connection_failed)
        self.cf.disconnected.add_callback(self._disconnected)

    def _remove_callbacks(self):
        def remove_callback(container, callback):
            try:
                container.remove_callback(callback)
            except ValueError:
                pass

        remove_callback(self.cf.connected, self._connected)
        remove_callback(self.cf.connection_failed, self._connection_failed)
        remove_callback(self.cf.disconnected, self._disconnected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
Fake link driver used to debug the UI without using the Crazyflie.

The operation of this driver can be controlled in two ways, either by
connecting to different URIs or by sending messages to the DebugDriver port
though CRTP once connected.

For normal connections a console thread is also started that will send
generated console output via CRTP.
"""
import errno
import logging
import queue
import random
import re
import string
import struct
import time
from datetime import datetime
from threading import Thread

from .crtpdriver import CRTPDriver
from .crtpstack import CRTPPacket
from .crtpstack import CRTPPort
from .exceptions import WrongUriType
from cflib.crazyflie.log import LogTocElement
from cflib.crazyflie.param import ParamTocElement

__author__ = 'Bitcraze AB'
__all__ = ['DebugDriver']

logger = logging.getLogger(__name__)

# This setup is used to debug raw memory logging
memlogging = {0x01: {'min': 0, 'max': 255, 'mod': 1, 'vartype': 1},
              0x02: {'min': 0, 'max': 65000, 'mod': 100, 'vartype': 2},
              0x03: {'min': 0, 'max': 100000, 'mod': 1000, 'vartype': 3},
              0x04: {'min': -100, 'max': 100, 'mod': 1, 'vartype': 4},
              0x05: {'min': -10000, 'max': 10000, 'mod': 2000, 'vartype': 5},
              0x06: {'min': -50000, 'max': 50000, 'mod': 1000, 'vartype': 6},
              0x07: {'min': 0, 'max': 255, 'mod': 1, 'vartype': 1}}


class FakeMemory:
    TYPE_I2C = 0
    TYPE_1W = 1

    def __init__(self, type, size, addr, data=None):
        self.type = type
        self.size = size
        self.addr = addr
        self.data = [0] * size
        if data:
            for i in range(len(data)):
                self.data[i] = data[i]

    def erase(self):
        self.data = [0] * self.size


class DebugDriver(CRTPDriver):
    """ Debug driver used for debugging UI/communication without using a
    Crazyflie"""

    def __init__(self):
        self.fakeLoggingThreads = []
        self._fake_mems = []
        self.needs_resending = False
        # Fill up the fake logging TOC with values and data
        self.fakeLogToc = []
        self.fakeLogToc.append({'varid': 0, 'vartype': 5, 'vargroup': 'imu',
                                'varname': 'gyro_x', 'min': -10000,
                                'max': 10000, 'mod': 1000})
        self.fakeLogToc.append({'varid': 1, 'vartype': 5, 'vargroup': 'imu',
                                'varname': 'gyro_y', 'min': -10000,
                                'max': 10000, 'mod': 150})
        self.fakeLogToc.append({'varid': 2, 'vartype': 5, 'vargroup': 'imu',
                                'varname': 'gyro_z', 'min': -10000,
                                'max': 10000, 'mod': 200})
        self.fakeLogToc.append({'varid': 3, 'vartype': 5, 'vargroup': 'imu',
                                'varname': 'acc_x', 'min': -1000,
                                'max': 1000, 'mod': 15})
        self.fakeLogToc.append({'varid': 4, 'vartype': 5, 'vargroup': 'imu',
                                'varname': 'acc_y', 'min': -1000,
                                'max': 1000, 'mod': 10})
        self.fakeLogToc.append({'varid': 5, 'vartype': 5, 'vargroup': 'imu',
                                'varname': 'acc_z', 'min': -1000,
                                'max': 1000, 'mod': 20})
        self.fakeLogToc.append({'varid': 6, 'vartype': 7,
                                'vargroup': 'stabilizer', 'varname': 'roll',
                                'min': -90, 'max': 90, 'mod': 2})
        self.fakeLogToc.append({'varid': 7, 'vartype': 7,
                                'vargroup': 'stabilizer', 'varname': 'pitch',
                                'min': -90, 'max': 90, 'mod': 1.5})
        self.fakeLogToc.append({'varid': 8, 'vartype': 7,
                                'vargroup': 'stabilizer', 'varname': 'yaw',
                                'min': -90, 'max': 90, 'mod': 2.5})
        self.fakeLogToc.append({'varid': 9, 'vartype': 7, 'vargroup': 'pm',
                                'varname': 'vbat', 'min': 3.0,
                                'max': 4.2, 'mod': 0.1})
        self.fakeLogToc.append({'varid': 10, 'vartype': 6, 'vargroup': 'motor',
                                'varname': 'm1', 'min': 0, 'max': 65000,
                                'mod': 1000})
        self.fakeLogToc.append({'varid': 11, 'vartype': 6, 'vargroup': 'motor',
                                'varname': 'm2', 'min': 0, 'max': 65000,
                                'mod': 1000})
        self.fakeLogToc.append({'varid': 12, 'vartype': 6, 'vargroup': 'motor',
                                'varname': 'm3', 'min': 0, 'max': 65000,
                                'mod': 1000})
        self.fakeLogToc.append({'varid': 13, 'vartype': 6, 'vargroup': 'motor',
                                'varname': 'm4', 'min': 0, 'max': 65000,
                                'mod': 1000})
        self.fakeLogToc.append({'varid': 14, 'vartype': 2,
                                'vargroup': 'stabilizer', 'varname': 'thrust',
                                'min': 0, 'max': 65000, 'mod': 1000})
        self.fakeLogToc.append({'varid': 15, 'vartype': 7,
                                'vargroup': 'baro', 'varname': 'asl',
                                'min': 540, 'max': 545, 'mod': 0.5})
        self.fakeLogToc.append({'varid': 16, 'vartype': 7,
                                'vargroup': 'baro', 'varname': 'aslRaw',
                                'min': 540, 'max': 545, 'mod': 1.0})
        self.fakeLogToc.append({'varid': 17, 'vartype': 7,
                                'vargroup': 'posEstimatorAlt',
                                'varname': 'estimatedZ',
                                'min': 540, 'max': 545, 'mod': 0.5})
        self.fakeLogToc.append({'varid': 18, 'vartype': 7,
                                'vargroup': 'baro', 'varname': 'temp',
                                'min': 26, 'max': 38, 'mod': 1.0})
        self.fakeLogToc.append({'varid': 19, 'vartype': 7,
                                'vargroup': 'posCtlAlt',
                                'varname': 'targetZ',
                                'min': 542, 'max': 543, 'mod': 0.1})
        self.fakeLogToc.append({'varid': 20, 'vartype': 6,
                                'vargroup': 'gps', 'varname': 'lat',
                                'min': 556112190, 'max': 556112790,
                                'mod': 10})
        self.fakeLogToc.append({'varid': 21, 'vartype': 6,
                                'vargroup': 'gps', 'varname': 'lon',
                                'min': 129945110, 'max': 129945710,
                                'mod': 10})
        self.fakeLogToc.append({'varid': 22, 'vartype': 6,
                                'vargroup': 'gps', 'varname': 'hMSL',
                                'min': 0, 'max': 100000,
                                'mod': 1000})
        self.fakeLogToc.append({'varid': 23, 'vartype': 6,
                                'vargroup': 'gps', 'varname': 'heading',
                                'min': -10000000, 'max': 10000000,
                                'mod': 100000})
        self.fakeLogToc.append({'varid': 24, 'vartype': 6,
                                'vargroup': 'gps', 'varname': 'gSpeed',
                                'min': 0, 'max': 1000,
                                'mod': 100})
        self.fakeLogToc.append({'varid': 25, 'vartype': 3,
                                'vargroup': 'gps', 'varname': 'hAcc',
                                'min': 0, 'max': 5000,
                                'mod': 100})
        self.fakeLogToc.append({'varid': 26, 'vartype': 1,
                                'vargroup': 'gps', 'varname': 'fixType',
                                'min': 0, 'max': 5,
                                'mod': 1})

        # Fill up the fake logging TOC with values and data
        self.fakeParamToc = []
        self.fakeParamToc.append({'varid': 0, 'vartype': 0x08,
                                  'vargroup': 'blah', 'varname': 'p',
                                  'writable': True, 'value': 100})
        self.fakeParamToc.append({'varid': 1, 'vartype': 0x0A,
                                  'vargroup': 'info', 'varname': 'cid',
                                  'writable': False, 'value': 1234})
        self.fakeParamToc.append({'varid': 2, 'vartype': 0x06,
                                  'vargroup': 'rpid', 'varname': 'prp',
                                  'writable': True, 'value': 1.5})
        self.fakeParamToc.append({'varid': 3, 'vartype': 0x06,
                                  'vargroup': 'rpid', 'varname': 'pyaw',
                                  'writable': True, 'value': 2.5})
        self.fakeParamToc.append({'varid': 4, 'vartype': 0x06,
                                  'vargroup': 'rpid', 'varname': 'irp',
                                  'writable': True, 'value': 3.5})
        self.fakeParamToc.append({'varid': 5, 'vartype': 0x06,
                                  'vargroup': 'rpid', 'varname': 'iyaw',
                                  'writable': True, 'value': 4.5})
        self.fakeParamToc.append({'varid': 6, 'vartype': 0x06,
                                  'vargroup': 'pid_attitude',
                                  'varname': 'pitch_kd', 'writable': True,
                                  'value': 5.5})
        self.fakeParamToc.append({'varid': 7, 'vartype': 0x06,
                                  'vargroup': 'rpid', 'varname': 'dyaw',
                                  'writable': True, 'value': 6.5})
        self.fakeParamToc.append({'varid': 8, 'vartype': 0x06,
                                  'vargroup': 'apid', 'varname': 'prp',
                                  'writable': True, 'value': 7.5})
        self.fakeParamToc.append({'varid': 9, 'vartype': 0x06,
                                  'vargroup': 'apid', 'varname': 'pyaw',
                                  'writable': True, 'value': 8.5})
        self.fakeParamToc.append({'varid': 10, 'vartype': 0x06,
                                  'vargroup': 'apid', 'varname': 'irp',
                                  'writable': True, 'value': 9.5})
        self.fakeParamToc.append({'varid': 11, 'vartype': 0x06,
                                  'vargroup': 'apid', 'varname': 'iyaw',
                                  'writable': True, 'value': 10.5})
        self.fakeParamToc.append({'varid': 12, 'vartype': 0x06,
                                  'vargroup': 'apid', 'varname': 'drp',
                                  'writable': True, 'value': 11.5})
        self.fakeParamToc.append({'varid': 13, 'vartype': 0x06,
                                  'vargroup': 'apid', 'varname': 'dyaw',
                                  'writable': True, 'value': 12.5})
        self.fakeParamToc.append({'varid': 14, 'vartype': 0x08,
                                  'vargroup': 'flightctrl',
                                  'varname': 'xmode', 'writable': True,
                                  'value': 1})
        self.fakeParamToc.append({'varid': 15, 'vartype': 0x08,
                                  'vargroup': 'flightctrl',
                                  'varname': 'ratepid', 'writable': True,
                                  'value': 1})
        self.fakeParamToc.append({'varid': 16, 'vartype': 0x08,
                                  'vargroup': 'imu_sensors',
                                  'varname': 'HMC5883L', 'writable': False,
                                  'value': 1})
        self.fakeParamToc.append({'varid': 17, 'vartype': 0x08,
                                  'vargroup': 'imu_sensors',
                                  'varname': 'MS5611', 'writable': False,
                                  'value': 1})
        self.fakeParamToc.append({'varid': 18, 'vartype': 0x0A,
                                  'vargroup': 'firmware',
                                  'varname': 'revision0', 'writable': False,
                                  'value': 0xdeb})
        self.fakeParamToc.append({'varid': 19, 'vartype': 0x09,
                                  'vargroup': 'firmware',
                                  'varname': 'revision1', 'writable': False,
                                  'value': 0x99})
        self.fakeParamToc.append({'varid': 20, 'vartype': 0x08,
                                  'vargroup': 'firmware',
                                  'varname': 'modified', 'writable': False,
                                  'value': 1})
        self.fakeParamToc.append({'varid': 21, 'vartype': 0x08,
                                  'vargroup': 'imu_tests',
                                  'varname': 'MPU6050', 'writable': False,
                                  'value': 1})
        self.fakeParamToc.append({'varid': 22, 'vartype': 0x08,
                                  'vargroup': 'imu_tests',
                                  'varname': 'HMC5883L', 'writable': False,
                                  'value': 1})
        self.fakeParamToc.append({'varid': 23, 'vartype': 0x08,
                                  'vargroup': 'imu_tests',
                                  'varname': 'MS5611', 'writable': False,
                                  'value': 1})

        self.fakeflash = {}
        self._random_answer_delay = True
        self.queue = queue.Queue()
        self._packet_handler = _PacketHandlingThread(self.queue,
                                                     self.fakeLogToc,
                                                     self.fakeParamToc,
                                                     self._fake_mems)
        self._packet_handler.start()

    def scan_interface(self, address):
        return [['debug://0/0', 'Normal connection'],
                ['debug://0/1', 'Fail to connect'],
                ['debug://0/2', 'Incomplete log TOC download'],
                ['debug://0/3', 'Insert random delays on replies'],
                ['debug://0/4',
                 'Insert random delays on replies and random TOC CRCs'],
                ['debug://0/5', 'Normal but random TOC CRCs'],
                ['debug://0/6', 'Normal but empty I2C and OW mems']]

    def get_status(self):
        return 'Ok'

    def get_name(self):
        return 'debug'

    def connect(self, uri, linkQualityCallback, linkErrorCallback):

        if not re.search('^debug://', uri):
            raise WrongUriType('Not a debug URI')

        self._packet_handler.linkErrorCallback = linkErrorCallback
        self._packet_handler.linkQualityCallback = linkQualityCallback

        # Debug-options for this driver that
        # is set by using different connection URIs
        self._packet_handler.inhibitAnswers = False
        self._packet_handler.doIncompleteLogTOC = False
        self._packet_handler.bootloader = False
        self._packet_handler._random_answer_delay = False
        self._packet_handler._random_toc_crcs = False

        if (re.search('^debug://.*/1$', uri)):
            self._packet_handler.inhibitAnswers = True
        if (re.search('^debug://.*/110$', uri)):
            self._packet_handler.bootloader = True
        if (re.search('^debug://.*/2$', uri)):
            self._packet_handler.doIncompleteLogTOC = True
        if (re.search('^debug://.*/3$', uri)):
            self._packet_handler._random_answer_delay = True
        if (re.search('^debug://.*/4$', uri)):
            self._packet_handler._random_answer_delay = True
            self._packet_handler._random_toc_crcs = True
        if (re.search('^debug://.*/5$', uri)):
            self._packet_handler._random_toc_crcs = True

        if len(self._fake_mems) == 0:
            # Add empty EEPROM
            self._fake_mems.append(FakeMemory(type=0, size=100, addr=0))
            # Add EEPROM with settings
            self._fake_mems.append(
                FakeMemory(type=0, size=100, addr=0,
                           data=[48, 120, 66, 67, 1, 8, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 231, 8, 231, 231, 231, 218]))
            # Add 1-wire memory with settings for LED-ring
            self._fake_mems.append(
                FakeMemory(type=1, size=112, addr=0x1234567890ABCDEF,
                           data=[0xeb, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01,
                                 0x44, 0x00, 0x0e,
                                 0x01, 0x09, 0x62, 0x63, 0x4c, 0x65, 0x64,
                                 0x52, 0x69, 0x6e,
                                 0x67, 0x02, 0x01, 0x62, 0x55]))
            # Add 1-wire memory with settings for LED-ring but bad CRC
            self._fake_mems.append(
                FakeMemory(type=1, size=112, addr=0x1234567890ABCDEF,
                           data=[0xeb, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01,
                                 0x44, 0x00, 0x0e,
                                 0x01, 0x09, 0x62, 0x63, 0x4c, 0x65, 0x64,
                                 0x52, 0x69, 0x6e,
                                 0x67, 0x02, 0x01, 0x62, 0x56]))
            # Add empty 1-wire memory
            self._fake_mems.append(
                FakeMemory(type=1, size=112, addr=0x1234567890ABCDEE,
                           data=[0x00 for a in range(112)]))

        if (re.search('^debug://.*/6$', uri)):
            logger.info('------------->Erasing memories on connect')
            for m in self._fake_mems:
                m.erase()

        self.fakeConsoleThread = None

        if (not self._packet_handler.inhibitAnswers and
                not self._packet_handler.bootloader):
            self.fakeConsoleThread = FakeConsoleThread(self.queue)
            self.fakeConsoleThread.start()

        if (self._packet_handler.linkQualityCallback is not None):
            self._packet_handler.linkQualityCallback(0)

    def receive_packet(self, time=0):
        if time == 0:
            try:
                return self.queue.get(False)
            except queue.Empty:
                return None
        elif time < 0:
            try:
                return self.queue.get(True)
            except queue.Empty:
                return None
        else:
            try:
                return self.queue.get(True, time)
            except queue.Empty:
                return None

    def send_packet(self, pk):
        self._packet_handler.handle_packet(pk)

    def close(self):
        logger.info('Closing debugdriver')
        for f in self._packet_handler.fakeLoggingThreads:
            f.stop()
        if self.fakeConsoleThread:
            self.fakeConsoleThread.stop()


class _PacketHandlingThread(Thread):
    """Thread for handling packets asynchronously"""

    def __init__(self, out_queue, fake_log_toc, fake_param_toc, fake_mems):
        Thread.__init__(self)
        self.setDaemon(True)
        self.queue = out_queue
        self.fakeLogToc = fake_log_toc
        self.fakeParamToc = fake_param_toc
        self._fake_mems = fake_mems
        self._in_queue = queue.Queue()

        self.inhibitAnswers = False
        self.doIncompleteLogTOC = False
        self.bootloader = False
        self._random_answer_delay = False
        self._random_toc_crcs = False

        self.linkErrorCallback = None
        self.linkQualityCallback = None
        random.seed(None)
        self.fakeLoggingThreads = []

        self._added_blocks = []

        self.nowAnswerCounter = 4

    def handle_packet(self, pk):
        self._in_queue.put(pk)

    def run(self):
        while (True):
            pk = self._in_queue.get(True)
            if (self.inhibitAnswers):
                self.nowAnswerCounter = self.nowAnswerCounter - 1
                logger.debug(
                    'Not answering with any data, will send link errori'
                    ' in %d retries', self.nowAnswerCounter)
                if (self.nowAnswerCounter == 0):
                    self.linkErrorCallback('Nothing is answering, and it'
                                           " shouldn't")
            else:
                if (pk.port == 0xFF):
                    self._handle_bootloader(pk)
                elif (pk.port == CRTPPort.DEBUGDRIVER):
                    self._handle_debugmessage(pk)
                elif (pk.port == CRTPPort.COMMANDER):
                    pass
                elif (pk.port == CRTPPort.LOGGING):
                    self._handle_logging(pk)
                elif (pk.port == CRTPPort.PARAM):
                    self.handleParam(pk)
                elif (pk.port == CRTPPort.MEM):
                    self._handle_mem_access(pk)
                else:
                    logger.warning(
                        'Not handling incoming packets on port [%d]',
                        pk.port)

    def _handle_mem_access(self, pk):
        chan = pk.channel
        cmd = pk.data[0]
        payload = pk.data[1:]

        if chan == 0:  # Info channel
            p_out = CRTPPacket()
            p_out.set_header(CRTPPort.MEM, 0)
            if cmd == 1:  # Request number of memories
                p_out.data = (1, len(self._fake_mems))
            if cmd == 2:
                id = payload[0]
                logger.info('Getting mem {}'.format(id))
                m = self._fake_mems[id]
                p_out.data = struct.pack(
                    '<BBBIQ', 2, id, m.type, m.size, m.addr)
            self._send_packet(p_out)

        if chan == 1:  # Read channel
            id = cmd
            addr = struct.unpack('I', payload[0:4])[0]
            length = payload[4]
            status = 0
            logger.info('MEM: Read {}bytes at 0x{:X} from memory {}'.format(
                length, addr, id))
            m = self._fake_mems[id]
            p_out = CRTPPacket()
            p_out.set_header(CRTPPort.MEM, 1)
            p_out.data = struct.pack('<BIB', id, addr, status)
            p_out.data += struct.pack('B' * length,
                                      *m.data[addr:addr + length])
            self._send_packet(p_out)

        if chan == 2:  # Write channel
            id = cmd
            addr = struct.unpack('I', payload[0:4])[0]
            data = payload[4:]
            logger.info('MEM: Write {}bytes at 0x{:X} to memory {}'.format(
                len(data), addr, id))
            m = self._fake_mems[id]

            for i in range(len(data)):
                m.data[addr + i] = data[i]

            status = 0

            p_out = CRTPPacket()
            p_out.set_header(CRTPPort.MEM, 2)
            p_out.data = struct.pack('<BIB', id, addr, status)
            self._send_packet(p_out)

    def _handle_bootloader(self, pk):
        cmd = pk.data[1]
        if (cmd == 0x10):  # Request info about copter
            p = CRTPPacket()
            p.set_header(0xFF, 0xFF)
            pageSize = 1024
            buffPages = 10
            flashPages = 100
            flashStart = 1
            p.data = struct.pack('<BBHHHH', 0xFF, 0x10, pageSize, buffPages,
                                 flashPages, flashStart)
            p.data += struct.pack('B' * 12, 0xA0A1A2A3A4A5)
            self._send_packet(p)
            logging.info('Bootloader: Sending info back info')
        elif (cmd == 0x14):  # Upload buffer
            [page, addr] = struct.unpack('<HH', p.data[0:4])
        elif (cmd == 0x18):  # Flash page
            p = CRTPPacket()
            p.set_header(0xFF, 0xFF)
            p.data = struct.pack('<BBH', 0xFF, 0x18, 1)
            self._send_packet(p)
        elif (cmd == 0xFF):  # Reset to firmware
            logger.info('Bootloader: Got reset command')
        else:
            logger.warning('Bootloader: Unknown command 0x%02X', cmd)

    def _handle_debugmessage(self, pk):
        if (pk.channel == 0):
            cmd = struct.unpack('B', pk.data[0])[0]
            if (cmd == 0):  # Fake link quality
                newLinkQuality = struct.unpack('B', pk.data[1])[0]
                self.linkQualityCallback(newLinkQuality)
            elif (cmd == 1):
                self.linkErrorCallback('DebugDriver was forced to disconnect!')
            else:
                logger.warning('Debug port: Not handling cmd=%d on channel 0',
                               cmd)
        else:
            logger.warning('Debug port: Not handling channel=%d',
                           pk.channel)

    def _handle_toc_access(self, pk):
        chan = pk.channel
        cmd = pk.data[0]
        logger.info('TOC access on port %d', pk.port)
        if (chan == 0):  # TOC Access
            cmd = pk.data[0]
            if (cmd == 0):  # Reqest variable info
                p = CRTPPacket()
                p.set_header(pk.port, 0)
                varIndex = 0
                if (len(pk.data) > 1):
                    varIndex = pk.data[1]
                    logger.debug('TOC[%d]: Requesting ID=%d', pk.port,
                                 varIndex)
                else:
                    logger.debug('TOC[%d]: Requesting first index..surprise,'
                                 ' it 0 !', pk.port)

                if (pk.port == CRTPPort.LOGGING):
                    entry = self.fakeLogToc[varIndex]
                if (pk.port == CRTPPort.PARAM):
                    entry = self.fakeParamToc[varIndex]

                vartype = entry['vartype']
                if (pk.port == CRTPPort.PARAM and entry['writable'] is True):
                    vartype = vartype | (0x10)

                p.data = struct.pack('<BBB', cmd, entry['varid'], vartype)
                for ch in entry['vargroup']:
                    p.data.append(ord(ch))
                p.data.append(0)
                for ch in entry['varname']:
                    p.data.append(ord(ch))
                p.data.append(0)
                if (self.doIncompleteLogTOC is False):
                    self._send_packet(p)
                elif (varIndex < 5):
                    self._send_packet(p)
                else:
                    logger.info('TOC: Doing incomplete TOC, stopping after'
                                ' varIndex => 5')

            if (cmd == 1):  # TOC CRC32 request
                fakecrc = 0
                if (pk.port == CRTPPort.LOGGING):
                    tocLen = len(self.fakeLogToc)
                    fakecrc = 0xAAAAAAAA
                if (pk.port == CRTPPort.PARAM):
                    tocLen = len(self.fakeParamToc)
                    fakecrc = 0xBBBBBBBB

                if self._random_toc_crcs:
                    fakecrc = int(''.join(
                        random.choice('ABCDEF' + string.digits) for x in
                        range(8)), 16)
                    logger.debug('Generated random TOC CRC: 0x%x', fakecrc)
                logger.info('TOC[%d]: Requesting TOC CRC, sending back fake'
                            ' stuff: %d', pk.port, len(self.fakeLogToc))
                p = CRTPPacket()
                p.set_header(pk.port, 0)
                p.data = struct.pack('<BBIBB', 1, tocLen, fakecrc, 16, 24)
                self._send_packet(p)

    def handleParam(self, pk):
        chan = pk.channel
        cmd = pk.data[0]
        logger.debug('PARAM: Port=%d, Chan=%d, cmd=%d', pk.port,
                     chan, cmd)
        if (chan == 0):  # TOC Access
            self._handle_toc_access(pk)
        elif (chan == 2):  # Settings access
            varId = pk.data[0]
            formatStr = ParamTocElement.types[
                self.fakeParamToc[varId]['vartype']][1]
            newvalue = struct.unpack(formatStr, pk.data[1:])[0]
            self.fakeParamToc[varId]['value'] = newvalue
            logger.info('PARAM: New value [%s] for param [%d]', newvalue,
                        varId)
            # Send back the new value
            p = CRTPPacket()
            p.set_header(pk.port, 2)
            p.data += struct.pack('<B', varId)
            p.data += struct.pack(formatStr, self.fakeParamToc[varId]['value'])
            self._send_packet(p)
        elif (chan == 1):
            p = CRTPPacket()
            p.set_header(pk.port, 1)
            varId = cmd
            p.data.append(varId)
            formatStr = ParamTocElement.types[
                self.fakeParamToc[varId]['vartype']][1]
            p.data += struct.pack(formatStr, self.fakeParamToc[varId]['value'])
            logger.info('PARAM: Getting value for %d', varId)
            self._send_packet(p)

    def _handle_logging(self, pk):
        chan = pk.channel
        cmd = pk.data[0]
        logger.debug('LOG: Chan=%d, cmd=%d', chan, cmd)
        if (chan == 0):  # TOC Access
            self._handle_toc_access(pk)
        elif (chan == 1):  # Settings access
            if (cmd == 0):
                blockId = pk.data[1]
                if blockId not in self._added_blocks:
                    self._added_blocks.append(blockId)
                    logger.info('LOG:Adding block id=%d', blockId)
                    listofvars = pk.data[3:]
                    fakeThread = _FakeLoggingDataThread(self.queue, blockId,
                                                        listofvars,
                                                        self.fakeLogToc)
                    self.fakeLoggingThreads.append(fakeThread)
                    fakeThread.start()
                    # Anser that everything is ok
                    p = CRTPPacket()
                    p.set_header(5, 1)
                    p.data = struct.pack('<BBB', 0, blockId, 0x00)
                    self._send_packet(p)
                else:
                    p = CRTPPacket()
                    p.set_header(5, 1)
                    p.data = struct.pack('<BBB', 0, blockId, errno.EEXIST)
                    self._send_packet(p)
            if (cmd == 1):
                logger.warning('LOG: Appending block not implemented!')
            if (cmd == 2):
                blockId = pk.data[1]
                logger.info('LOG: Should delete block %d', blockId)
                success = False
                for fb in self.fakeLoggingThreads:
                    if (fb.blockId == blockId):
                        fb._disable_logging()
                        fb.stop()

                        p = CRTPPacket()
                        p.set_header(5, 1)
                        p.data = struct.pack('<BBB', cmd, blockId, 0x00)
                        self._send_packet(p)
                        logger.info('LOG: Deleted block=%d', blockId)
                        success = True
                if (success is False):
                    logger.warning('LOG: Could not delete block=%d, not found',
                                   blockId)
                    # TODO: Send back error code

            if (cmd == 3):
                blockId = pk.data[1]
                period = pk.data[2] * 10  # Sent as multiple of 10 ms
                logger.info('LOG:Starting block %d', blockId)
                success = False
                for fb in self.fakeLoggingThreads:
                    if (fb.blockId == blockId):
                        fb._enable_logging()
                        fb.period = period
                        p = CRTPPacket()
                        p.set_header(5, 1)
                        p.data = struct.pack('<BBB', cmd, blockId, 0x00)
                        self._send_packet(p)
                        logger.info('LOG:Started block=%d', blockId)
                        success = True
                if (success is False):
                    logger.info('LOG:Could not start block=%d, not found',
                                blockId)
                    # TODO: Send back error code
            if (cmd == 4):
                blockId = pk.data[1]
                logger.info('LOG:Pausing block %d', blockId)
                success = False
                for fb in self.fakeLoggingThreads:
                    if (fb.blockId == blockId):
                        fb._disable_logging()
                        p = CRTPPacket()
                        p.set_header(5, 1)
                        p.data = struct.pack('<BBB', cmd, blockId, 0x00)
                        self._send_packet(p)
                        logger.info('LOG:Pause block=%d', blockId)
                        success = True
                if (success is False):
                    logger.warning('LOG:Could not pause block=%d, not found',
                                   blockId)
                    # TODO: Send back error code
            if (cmd == 5):
                logger.info('LOG: Reset logging, but doing nothing')
                p = CRTPPacket()
                p.set_header(5, 1)
                p.data = struct.pack('<BBB', cmd, 0x00, 0x00)
                self._send_packet(p)

        elif (chan > 1):
            logger.warning('LOG: Uplink packets with channels > 1 not'
                           ' supported!')

    def _send_packet(self, pk):
        # Do not delay log data
        if (self._random_answer_delay and pk.port != 0x05 and
                pk.channel != 0x02):
            # Calculate a delay between 0ms and 250ms
            delay = random.randint(0, 250) / 1000.0
            logger.debug('Delaying answer %.2fms', delay * 1000)
            time.sleep(delay)
        self.queue.put(pk)


class _FakeLoggingDataThread(Thread):
    """Thread that will send back fake logging data via CRTP"""

    def __init__(self, outQueue, blockId, listofvars, fakeLogToc):
        Thread.__init__(self)
        self.starttime = datetime.now()
        self.outQueue = outQueue
        self.setDaemon(True)
        self.mod = 0
        self.blockId = blockId
        self.period = 0
        self.listofvars = listofvars
        self.shouldLog = False
        self.fakeLogToc = fakeLogToc
        self.fakeLoggingData = []
        self.setName('Fakelog block=%d' % blockId)
        self.shouldQuit = False

        logging.info('FakeDataLoggingThread created for blockid=%d', blockId)
        i = 0
        while (i < len(listofvars)):
            varType = listofvars[i]
            var_stored_as = (varType >> 8)
            var_fetch_as = (varType & 0xFF)
            if (var_stored_as > 0):
                addr = struct.unpack('<I', listofvars[i + 1:i + 5])
                logger.debug('FakeLoggingThread: We should log a memory addr'
                             ' 0x%04X', addr)
                self.fakeLoggingData.append([memlogging[var_fetch_as],
                                             memlogging[var_fetch_as]['min'],
                                             1])
                i = i + 5
            else:
                varId = listofvars[i]
                logger.debug('FakeLoggingThread: We should log variable from'
                             ' TOC: id=%d, type=0x%02X', varId, varType)
                for t in self.fakeLogToc:
                    if (varId == t['varid']):
                        # Each touple will have var data and current fake value
                        self.fakeLoggingData.append([t, t['min'], 1])
                i = i + 2

    def _enable_logging(self):
        self.shouldLog = True
        logging.info('_FakeLoggingDataThread: Enable thread [%s] at period %d',
                     self.getName(), self.period)

    def _disable_logging(self):
        self.shouldLog = False
        logging.info('_FakeLoggingDataThread: Disable thread [%s]',
                     self.getName())

    def stop(self):
        self.shouldQuit = True

    def run(self):
        while (self.shouldQuit is False):
            if (self.shouldLog is True):

                p = CRTPPacket()
                p.set_header(5, 2)
                p.data = struct.pack('<B', self.blockId)
                timestamp = int(
                    (datetime.now() - self.starttime).total_seconds() * 1000)
                p.data += struct.pack('BBB', timestamp & 0xff,
                                      (timestamp >> 8) & 0x0ff,
                                      (timestamp >> 16) & 0x0ff)  # Timestamp

                for d in self.fakeLoggingData:
                    # Set new value
                    d[1] = d[1] + d[0]['mod'] * d[2]
                    # Obej the limitations
                    if (d[1] > d[0]['max']):
                        d[1] = d[0]['max']  # Limit value
                        d[2] = -1  # Switch direction
                    if (d[1] < d[0]['min']):
                        d[1] = d[0]['min']  # Limit value
                        d[2] = 1  # Switch direction
                    # Pack value
                    formatStr = LogTocElement.types[d[0]['vartype']][1]
                    p.data += struct.pack(formatStr, d[1])
                self.outQueue.put(p)
            time.sleep(self.period / 1000.0)  # Period in ms here


class FakeConsoleThread(Thread):
    """Thread that will send back fake console data via CRTP"""

    def __init__(self, outQueue):
        Thread.__init__(self)
        self.outQueue = outQueue
        self.setDaemon(True)
        self._should_run = True

    def stop(self):
        self._shoud_run = False

    def run(self):
        # Temporary hack to test GPS from firmware by sending NMEA string
        # on console
        long_val = 0
        lat_val = 0
        alt_val = 0

        while (self._should_run):
            long_val += 1
            lat_val += 1
            alt_val += 1.0

            long_string = '5536.677%d' % (long_val % 99)
            lat_string = '01259.645%d' % (lat_val % 99)
            alt_string = '%.1f' % (alt_val % 100.0)

            # Copy of what is sent from the module, but note that only
            # the GPGGA message is being simulated, the others are fixed...
            self._send_text('Time is now %s\n' % datetime.now())
            self._send_text('$GPVTG,,T,,M,0.386,N,0.716,K,A*2E\n')
            self._send_text('$GPGGA,135544.0')
            self._send_text('0,%s,N,%s,E,1,04,2.62,3.6,M,%s,M,,*58\n' % (
                long_string, lat_string, alt_string))
            self._send_text(
                '$GPGSA,A,3,31,20,23,07,,,,,,,,,3.02,2.62,1.52*05\n')
            self._send_text('$GPGSV,2,1,07,07,09,181,15,13,63,219,26,16,02,'
                            '097,,17,05,233,20*7E\n')
            self._send_text(
                '$GPGSV,2,2,07,20,42,119,35,23,77,097,27,31,12,032,19*47\n')
            self._send_text(
                '$GPGLL,5536.67734,N,01259.64578,E,135544.00,A,A*68\n')

            time.sleep(2)

    def _send_text(self, message):
        p = CRTPPacket()
        p.set_header(0, 0)

        # This might be done prettier ;-)
        p.data = message

        self.outQueue.put(p)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
Enables reading/writing of parameter values to/from the Crazyflie.

When a Crazyflie is connected it's possible to download a TableOfContent of all
the parameters that can be written/read.

"""
import logging
import struct
from queue import Queue
from threading import Lock
from threading import Thread

from .toc import Toc
from .toc import TocFetcher
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller


__author__ = 'Bitcraze AB'
__all__ = ['Param', 'ParamTocElement']

logger = logging.getLogger(__name__)

# Possible states
IDLE = 0
WAIT_TOC = 1
WAIT_READ = 2
WAIT_WRITE = 3

TOC_CHANNEL = 0
READ_CHANNEL = 1
WRITE_CHANNEL = 2

# One element entry in the TOC


class ParamTocElement:
    """An element in the Log TOC."""

    RW_ACCESS = 0
    RO_ACCESS = 1

    types = {0x08: ('uint8_t', '<B'),
             0x09: ('uint16_t', '<H'),
             0x0A: ('uint32_t', '<L'),
             0x0B: ('uint64_t', '<Q'),
             0x00: ('int8_t', '<b'),
             0x01: ('int16_t', '<h'),
             0x02: ('int32_t', '<i'),
             0x03: ('int64_t', '<q'),
             0x05: ('FP16', ''),
             0x06: ('float', '<f'),
             0x07: ('double', '<d')}

    def __init__(self, ident=0, data=None):
        """TocElement creator. Data is the binary payload of the element."""
        self.ident = ident
        if (data):
            strs = struct.unpack('s' * len(data[1:]), data[1:])
            s = ''
            for ch in strs:
                s += ch.decode('ISO-8859-1')
            strs = s.split('\x00')
            self.group = strs[0]
            self.name = strs[1]

            metadata = data[0]
            if type(metadata) == str:
                metadata = ord(metadata)

            self.ctype = self.types[metadata & 0x0F][0]
            self.pytype = self.types[metadata & 0x0F][1]
            if ((metadata & 0x40) != 0):
                self.access = ParamTocElement.RO_ACCESS
            else:
                self.access = ParamTocElement.RW_ACCESS

    def get_readable_access(self):
        if (self.access == ParamTocElement.RO_ACCESS):
            return 'RO'
        return 'RW'


class Param():
    """
    Used to read and write parameter values in the Crazyflie.
    """

    def __init__(self, crazyflie):
        self.toc = Toc()

        self.cf = crazyflie
        self._useV2 = False
        self.param_update_callbacks = {}
        self.group_update_callbacks = {}
        self.all_update_callback = Caller()
        self.param_updater = None

        self.param_updater = _ParamUpdater(
            self.cf, self._useV2, self._param_updated)
        self.param_updater.start()

        self.cf.disconnected.add_callback(self._disconnected)

        self.all_updated = Caller()
        self.is_updated = False

        self.values = {}

    def request_update_of_all_params(self):
        """Request an update of all the parameters in the TOC"""
        for group in self.toc.toc:
            for name in self.toc.toc[group]:
                complete_name = '%s.%s' % (group, name)
                self.request_param_update(complete_name)

    def _check_if_all_updated(self):
        """Check if all parameters from the TOC has at least been fetched
        once"""
        for g in self.toc.toc:
            if g not in self.values:
                return False
            for n in self.toc.toc[g]:
                if n not in self.values[g]:
                    return False

        return True

    def _param_updated(self, pk):
        """Callback with data for an updated parameter"""
        if self._useV2:
            var_id = struct.unpack('<H', pk.data[:2])[0]
        else:
            var_id = pk.data[0]
        element = self.toc.get_element_by_id(var_id)
        if element:
            if self._useV2:
                s = struct.unpack(element.pytype, pk.data[2:])[0]
            else:
                s = struct.unpack(element.pytype, pk.data[1:])[0]
            s = s.__str__()
            complete_name = '%s.%s' % (element.group, element.name)

            # Save the value for synchronous access
            if element.group not in self.values:
                self.values[element.group] = {}
            self.values[element.group][element.name] = s

            logger.debug('Updated parameter [%s]' % complete_name)
            if complete_name in self.param_update_callbacks:
                self.param_update_callbacks[complete_name].call(
                    complete_name, s)
            if element.group in self.group_update_callbacks:
                self.group_update_callbacks[element.group].call(
                    complete_name, s)
            self.all_update_callback.call(complete_name, s)

            # Once all the parameters are updated call the
            # callback for "everything updated" (after all the param
            # updated callbacks)
            if self._check_if_all_updated() and not self.is_updated:
                self.is_updated = True
                self.all_updated.call()
        else:
            logger.debug('Variable id [%d] not found in TOC', var_id)

    def remove_update_callback(self, group, name=None, cb=None):
        """Remove the supplied callback for a group or a group.name"""
        if not cb:
            return

        if not name:
            if group in self.group_update_callbacks:
                self.group_update_callbacks[group].remove_callback(cb)
        else:
            paramname = '{}.{}'.format(group, name)
            if paramname in self.param_update_callbacks:
                self.param_update_callbacks[paramname].remove_callback(cb)

    def add_update_callback(self, group=None, name=None, cb=None):
        """
        Add a callback for a specific parameter name. This callback will be
        executed when a new value is read from the Crazyflie.
        """
        if not group and not name:
            self.all_update_callback.add_callback(cb)
        elif not name:
            if group not in self.group_update_callbacks:
                self.group_update_callbacks[group] = Caller()
            self.group_update_callbacks[group].add_callback(cb)
        else:
            paramname = '{}.{}'.format(group, name)
            if paramname not in self.param_update_callbacks:
                self.param_update_callbacks[paramname] = Caller()
            self.param_update_callbacks[paramname].add_callback(cb)

    def refresh_toc(self, refresh_done_callback, toc_cache):
        """
        Initiate a refresh of the parameter TOC.
        """
        self._useV2 = self.cf.platform.get_protocol_version() >= 4
        toc_fetcher = TocFetcher(self.cf, ParamTocElement,
                                 CRTPPort.PARAM, self.toc,
                                 refresh_done_callback, toc_cache)
        toc_fetcher.start()

    def _disconnected(self, uri):
        """Disconnected callback from Crazyflie API"""
        self.param_updater.close()
        self.is_updated = False
        # Clear all values from the previous Crazyflie
        self.toc = Toc()
        self.values = {}

    def request_param_update(self, complete_name):
        """
        Request an update of the value for the supplied parameter.
        """
        self.param_updater.request_param_update(
            self.toc.get_element_id(complete_name))

    def set_value(self, complete_name, value):
        """
        Set the value for the supplied parameter.
        """
        element = self.toc.get_element_by_complete_name(complete_name)

        if not element:
            logger.warning("Cannot set value for [%s], it's not in the TOC!",
                           complete_name)
            raise KeyError('{} not in param TOC'.format(complete_name))
        elif element.access == ParamTocElement.RO_ACCESS:
            logger.debug('[%s] is read only, no trying to set value',
                         complete_name)
            raise AttributeError('{} is read-only!'.format(complete_name))
        else:
            varid = element.ident
            pk = CRTPPacket()
            pk.set_header(CRTPPort.PARAM, WRITE_CHANNEL)
            if self._useV2:
                pk.data = struct.pack('<H', varid)
            else:
                pk.data = struct.pack('<B', varid)

            try:
                value_nr = eval(value)
            except TypeError:
                value_nr = value

            pk.data += struct.pack(element.pytype, value_nr)
            self.param_updater.request_param_setvalue(pk)


class _ParamUpdater(Thread):
    """This thread will update params through a queue to make sure that we
    get back values"""

    def __init__(self, cf, useV2, updated_callback):
        """Initialize the thread"""
        Thread.__init__(self)
        self.setDaemon(True)
        self.wait_lock = Lock()
        self.cf = cf
        self._useV2 = useV2
        self.updated_callback = updated_callback
        self.request_queue = Queue()
        self.cf.add_port_callback(CRTPPort.PARAM, self._new_packet_cb)
        self._should_close = False
        self._req_param = -1

    def close(self):
        # First empty the queue from all packets
        while not self.request_queue.empty():
            self.request_queue.get()
        # Then force an unlock of the mutex if we are waiting for a packet
        # we didn't get back due to a disconnect for example.
        try:
            self.wait_lock.release()
        except Exception:
            pass

    def request_param_setvalue(self, pk):
        """Place a param set value request on the queue. When this is sent to
        the Crazyflie it will answer with the update param value. """
        self.request_queue.put(pk)

    def _new_packet_cb(self, pk):
        """Callback for newly arrived packets"""
        if pk.channel == READ_CHANNEL or pk.channel == WRITE_CHANNEL:
            if self._useV2:
                var_id = struct.unpack('<H', pk.data[:2])[0]
                if pk.channel == READ_CHANNEL:
                    pk.data = pk.data[:2] + pk.data[3:]
            else:
                var_id = pk.data[0]
            if (pk.channel != TOC_CHANNEL and self._req_param == var_id and
                    pk is not None):
                self.updated_callback(pk)
                self._req_param = -1
                try:
                    self.wait_lock.release()
                except Exception:
                    pass

    def request_param_update(self, var_id):
        """Place a param update request on the queue"""
        self._useV2 = self.cf.platform.get_protocol_version() >= 4
        pk = CRTPPacket()
        pk.set_header(CRTPPort.PARAM, READ_CHANNEL)
        if self._useV2:
            pk.data = struct.pack('<H', var_id)
        else:
            pk.data = struct.pack('<B', var_id)
        logger.debug('Requesting request to update param [%d]', var_id)
        self.request_queue.put(pk)

    def run(self):
        while not self._should_close:
            pk = self.request_queue.get()  # Wait for request update
            self.wait_lock.acquire()
            if self.cf.link:
                if self._useV2:
                    self._req_param = struct.unpack('<H', pk.data[:2])[0]
                    self.cf.send_packet(
                        pk, expected_reply=(tuple(pk.data[:2])))
                else:
                    self._req_param = pk.data[0]
                    self.cf.send_packet(
                        pk, expected_reply=(tuple(pk.data[:1])))
            else:
                self.wait_lock.release()



#!/usr/bin/env python3
from setuptools import find_packages
from setuptools import setup

setup(
    name='cflib',
    version='0.1.12.1',
    packages=find_packages(exclude=['examples', 'tests']),

    description='Crazyflie python driver',
    url='https://github.com/bitcraze/crazyflie-lib-python',

    author='Bitcraze and contributors',
    author_email='contact@bitcraze.io',
    license='GPLv3',

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ],

    keywords='driver crazyflie quadcopter',

    install_requires='pyusb>=1.0.0b2',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
Bootloading utilities for the Crazyflie.
"""
import json
import logging
import sys
import time
import zipfile

from .boottypes import BootVersion
from .boottypes import TargetTypes
from .cloader import Cloader

logger = logging.getLogger(__name__)

__author__ = 'Bitcraze AB'
__all__ = ['Bootloader']


class Bootloader:
    """Bootloader utility for the Crazyflie"""

    def __init__(self, clink=None):
        """Init the communication class by starting to communicate with the
        link given. clink is the link address used after resetting to the
        bootloader.

        The device is actually considered in firmware mode.
        """
        self.clink = clink
        self.in_loader = False

        self.page_size = 0
        self.buffer_pages = 0
        self.flash_pages = 0
        self.start_page = 0
        self.cpuid = 'N/A'
        self.error_code = 0
        self.protocol_version = 0

        # Outgoing callbacks for progress
        # int
        self.progress_cb = None
        # msg
        self.error_cb = None
        # bool
        self.in_bootloader_cb = None
        # Target
        self.dev_info_cb = None
        self.terminate_flashing_cb = None

        # self.dev_info_cb.add_callback(self._dev_info)
        # self.in_bootloader_cb.add_callback(self._bootloader_info)

        self._boot_plat = None

        self._cload = Cloader(clink,
                              info_cb=self.dev_info_cb,
                              in_boot_cb=self.in_bootloader_cb)

    def start_bootloader(self, warm_boot=False):
        if warm_boot:
            self._cload.open_bootloader_uri(self.clink)
            started = self._cload.reset_to_bootloader(TargetTypes.NRF51)
            if started:
                started = self._cload.check_link_and_get_info()
        else:
            uri = self._cload.scan_for_bootloader()

            # Workaround for libusb on Windows (open/close too fast)
            time.sleep(1)

            if uri:
                self._cload.open_bootloader_uri(uri)
                started = self._cload.check_link_and_get_info()
            else:
                started = False

        if started:
            self.protocol_version = self._cload.protocol_version

            if (self.protocol_version == BootVersion.CF1_PROTO_VER_0 or
                    self.protocol_version == BootVersion.CF1_PROTO_VER_1):
                # Nothing more to do
                pass
            elif self.protocol_version == BootVersion.CF2_PROTO_VER:
                self._cload.request_info_update(TargetTypes.NRF51)
            else:
                print('Bootloader protocol 0x{:X} not '
                      'supported!'.self.protocol_version)

        return started

    def get_target(self, target_id):
        return self._cload.request_info_update(target_id)

    def read_cf1_config(self):
        """Read a flash page from the specified target"""
        target = self._cload.targets[0xFF]
        config_page = target.flash_pages - 1

        return self._cload.read_flash(addr=0xFF, page=config_page)

    def write_cf1_config(self, data):
        target = self._cload.targets[0xFF]
        config_page = target.flash_pages - 1

        to_flash = {'target': target, 'data': data, 'type': 'CF1 config',
                    'start_page': config_page}

        self._internal_flash(target=to_flash)

    def flash(self, filename, targets):
        for target in targets:
            if TargetTypes.from_string(target) not in self._cload.targets:
                print('Target {} not found by bootloader'.format(target))
                return False

        files_to_flash = ()
        if zipfile.is_zipfile(filename):
            # Read the manifest (don't forget to check so there is one!)
            try:
                zf = zipfile.ZipFile(filename)
                js = zf.read('manifest.json').decode('UTF-8')
                j = json.loads(js)
                files = j['files']
                platform_id = self._get_platform_id()
                files_for_platform = self._filter_platform(files, platform_id)
                if len(targets) == 0:
                    targets = self._extract_targets_from_manifest_files(
                        files_for_platform)

                zip_targets = self._extract_zip_targets(files_for_platform)
            except KeyError as e:
                print(e)
                print('No manifest.json in {}'.format(filename))
                return

            try:
                # Match and create targets
                for target in targets:
                    t = targets[target]
                    for type in t:
                        file_to_flash = {}
                        current_target = '{}-{}'.format(target, type)
                        file_to_flash['type'] = type
                        # Read the data, if this fails we bail
                        file_to_flash['target'] = self._cload.targets[
                            TargetTypes.from_string(target)]
                        file_to_flash['data'] = zf.read(
                            zip_targets[target][type]['filename'])
                        file_to_flash['start_page'] = file_to_flash[
                            'target'].start_page
                        files_to_flash += (file_to_flash,)
            except KeyError:
                print('Could not find a file for {} in {}'.format(
                    current_target, filename))
                return False

        else:
            if len(targets) != 1:
                print('Not an archive, must supply one target to flash')
            else:
                file_to_flash = {}
                file_to_flash['type'] = 'binary'
                f = open(filename, 'rb')
                for t in targets:
                    file_to_flash['target'] = self._cload.targets[
                        TargetTypes.from_string(t)]
                    file_to_flash['type'] = targets[t][0]
                    file_to_flash['start_page'] = file_to_flash[
                        'target'].start_page
                file_to_flash['data'] = f.read()
                f.close()
                files_to_flash += (file_to_flash,)

        if not self.progress_cb:
            print('')

        file_counter = 0
        for target in files_to_flash:
            file_counter += 1
            self._internal_flash(target, file_counter, len(files_to_flash))

    def _filter_platform(self, files, platform_id):
        result = {}
        for file in files:
            file_info = files[file]
            file_platform = file_info['platform']
            if platform_id == file_platform:
                result[file] = file_info
        return result

    def _extract_zip_targets(self, files):
        zip_targets = {}
        for file in files:
            file_name = file
            file_info = files[file]
            file_target = file_info['target']
            file_type = file_info['type']
            if file_target not in zip_targets:
                zip_targets[file_target] = {}
            zip_targets[file_target][file_type] = {
                'filename': file_name}
        return zip_targets

    def _extract_targets_from_manifest_files(self, files):
        targets = {}
        for file in files:
            file_info = files[file]
            file_target = file_info['target']
            file_type = file_info['type']
            if file_target in targets:
                targets[file_target] += (file_type,)
            else:
                targets[file_target] = (file_type,)

        return targets

    def reset_to_firmware(self):
        if self._cload.protocol_version == BootVersion.CF2_PROTO_VER:
            self._cload.reset_to_firmware(TargetTypes.NRF51)
        else:
            self._cload.reset_to_firmware(TargetTypes.STM32)

    def close(self):
        if self._cload:
            self._cload.close()

    def _internal_flash(self, target, current_file_number=1, total_files=1):

        image = target['data']
        t_data = target['target']

        start_page = target['start_page']

        # If used from a UI we need some extra things for reporting progress
        factor = (100.0 * t_data.page_size) / len(image)
        progress = 0

        if self.progress_cb:
            self.progress_cb(
                '({}/{}) Starting...'.format(current_file_number, total_files),
                int(progress))
        else:
            sys.stdout.write(
                'Flashing {} of {} to {} ({}): '.format(
                    current_file_number, total_files,
                    TargetTypes.to_string(t_data.id), target['type']))
            sys.stdout.flush()

        if len(image) > ((t_data.flash_pages - start_page) *
                         t_data.page_size):
            if self.progress_cb:
                self.progress_cb(
                    'Error: Not enough space to flash the image file.',
                    int(progress))
            else:
                print('Error: Not enough space to flash the image file.')
            raise Exception()

        if not self.progress_cb:
            logger.info(('%d bytes (%d pages) ' % (
                (len(image) - 1), int(len(image) / t_data.page_size) + 1)))
            sys.stdout.write(('%d bytes (%d pages) ' % (
                (len(image) - 1), int(len(image) / t_data.page_size) + 1)))
            sys.stdout.flush()

        # For each page
        ctr = 0  # Buffer counter
        for i in range(0, int((len(image) - 1) / t_data.page_size) + 1):
            if self.terminate_flashing_cb and self.terminate_flashing_cb():
                raise Exception('Flashing terminated')

            # Load the buffer
            if ((i + 1) * t_data.page_size) > len(image):
                self._cload.upload_buffer(
                    t_data.addr, ctr, 0, image[i * t_data.page_size:])
            else:
                self._cload.upload_buffer(
                    t_data.addr, ctr, 0,
                    image[i * t_data.page_size: (i + 1) * t_data.page_size])

            ctr += 1

            if self.progress_cb:
                progress += factor
                self.progress_cb('({}/{}) Uploading buffer to {}...'.format(
                    current_file_number,
                    total_files,
                    TargetTypes.to_string(t_data.id)),

                    int(progress))
            else:
                sys.stdout.write('.')
                sys.stdout.flush()

            # Flash when the complete buffers are full
            if ctr >= t_data.buffer_pages:
                if self.progress_cb:
                    self.progress_cb('({}/{}) Writing buffer to {}...'.format(
                        current_file_number,
                        total_files,
                        TargetTypes.to_string(t_data.id)),

                        int(progress))
                else:
                    sys.stdout.write('%d' % ctr)
                    sys.stdout.flush()
                if not self._cload.write_flash(t_data.addr, 0,
                                               start_page + i - (ctr - 1),
                                               ctr):
                    if self.progress_cb:
                        self.progress_cb(
                            'Error during flash operation (code {})'.format(
                                self._cload.error_code),
                            int(progress))
                    else:
                        print('\nError during flash operation (code %d). '
                              'Maybe wrong radio link?' %
                              self._cload.error_code)
                    raise Exception()

                ctr = 0

        if ctr > 0:
            if self.progress_cb:
                self.progress_cb('({}/{}) Writing buffer to {}...'.format(
                    current_file_number,
                    total_files,
                    TargetTypes.to_string(t_data.id)),
                    int(progress))
            else:
                sys.stdout.write('%d' % ctr)
                sys.stdout.flush()
            if not self._cload.write_flash(
                    t_data.addr, 0,
                    (start_page + (int((len(image) - 1) / t_data.page_size)) -
                     (ctr - 1)), ctr):
                if self.progress_cb:
                    self.progress_cb(
                        'Error during flash operation (code {})'.format(
                            self._cload.error_code),
                        int(progress))
                else:
                    print('\nError during flash operation (code %d). Maybe'
                          ' wrong radio link?' % self._cload.error_code)
                raise Exception()

        if self.progress_cb:
            self.progress_cb(
                '({}/{}) Flashing done!'.format(current_file_number,
                                                total_files),
                int(100))
        else:
            print('')

    def _get_platform_id(self):
        """Get platform identifier used in the zip manifest for curr copter"""
        identifier = 'cf1'
        if (BootVersion.is_cf2(self.protocol_version)):
            identifier = 'cf2'

        return identifier
